import decimal
from _typeshed import ReadableBuffer
from datetime import date, datetime, time
from typing import Any, Sequence, overload
from typing_extensions import Literal

from .resultrow import ResultRow

apilevel: str
threadsafety: int
paramstyle: tuple[str, ...]
connect = Connection

class Connection:
    def __init__(
        self,
        address: str,
        port: int,
        username: str,
        password: str,
        autocommit: bool = ...,
        packetsize: int | None = ...,
        userkey: str | None = ...,
        *,
        sessionvariables: dict[str, str] | None = ...,
        forcebulkfetch: bool | None = ...,
    ) -> None: ...
    def cancel(self) -> bool: ...
    def close(self) -> None: ...
    def commit(self) -> None: ...
    def cursor(self) -> Cursor: ...
    def getaddress(self) -> str: ...
    def getautocommit(self) -> bool: ...
    def getclientinfo(self, key: str = ...) -> str | dict[str, str]: ...
    def isconnected(self) -> bool: ...
    def rollback(self) -> None: ...
    def setautocommit(self, auto: bool = ...) -> None: ...
    def setclientinfo(self, key: str, value: str | None = ...) -> None: ...

class LOB:
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...
    def close(self) -> bool: ...
    def find(self, object: str, length: int, position: int = ...) -> int: ...
    def read(self, size: int = ..., position: int = ...) -> str | bytes: ...
    def write(self, object: str | bytes) -> int: ...

_Parameters = Sequence[tuple[Any, ...]]

class Cursor:
    description: tuple[tuple[Any, ...], ...]
    rowcount: int
    statementhash: str | None
    connection: Connection
    arraysize: int
    def __init__(self, *args: Any, **kwargs: Any) -> None: ...
    def callproc(self, procname: str, parameters: tuple[Any, ...] = ..., overview: bool = ...) -> tuple[Any, ...]: ...
    def close(self) -> None: ...
    def description_ext(self) -> Sequence[tuple[Any, ...]]: ...
    def execute(self, operation: str, parameters: tuple[Any, ...]) -> bool: ...
    def executemany(self, operation: str, parameters: _Parameters) -> Any: ...
    def executemanyprepared(self, parameters: _Parameters) -> Any: ...
    def executeprepared(self, parameters: _Parameters = ...) -> Any: ...
    def fetchone(self, uselob: bool = ...) -> ResultRow | None: ...
    def fetchall(self) -> list[ResultRow]: ...
    def fetchmany(self, size: int | None = ...) -> list[ResultRow]: ...
    def get_resultset_holdability(self) -> int: ...
    def getwarning(self) -> Warning | None: ...
    def haswarning(self) -> bool: ...
    def nextset(self) -> None: ...
    def parameter_description(self) -> tuple[str, ...]: ...
    @overload
    def prepare(self, operation: str, newcursor: Literal[True]) -> Cursor: ...
    @overload
    def prepare(self, operation: str, newcursor: Literal[False]) -> Any: ...
    def scroll(self, value: int, mode: Literal["absolute", "relative"] = ...) -> None: ...
    def server_cpu_time(self) -> int: ...
    def server_memory_usage(self) -> int: ...
    def server_processing_time(self) -> int: ...
    def setinputsizes(self, *args: Any, **kwargs: Any) -> None: ...
    def setfetchsize(self, value: int) -> None: ...
    def set_resultset_holdability(self, holdability: int) -> None: ...
    def setoutputsize(self, *args: Any, **kwargs: Any) -> None: ...

class Warning(Exception):
    errorcode: int
    errortext: str

class Error(Exception):
    errorcode: int
    errortext: str

class DatabaseError(Error): ...
class OperationalError(DatabaseError): ...
class ProgrammingError(DatabaseError): ...
class IntegrityError(DatabaseError): ...
class InterfaceError(Error): ...
class InternalError(DatabaseError): ...
class DataError(DatabaseError): ...
class NotSupportedError(DatabaseError): ...

def Date(year: int, month: int, day: int) -> date: ...
def Time(hour: int, minute: int, second: int, millisecond: int = ...) -> time: ...
def Timestamp(year: int, month: int, day: int, hour: int, minute: int, second: int, millisecond: int = ...) -> datetime: ...
def DateFromTicks(ticks: float) -> date: ...
def TimeFromTicks(ticks: float) -> time: ...
def TimestampFromTicks(ticks: float) -> datetime: ...
def Binary(data: ReadableBuffer) -> memoryview: ...

Decimal = decimal.Decimal

NUMBER: type[int] | type[float] | type[complex]
DATETIME: type[date] | type[time] | type[datetime]
STRING = str
BINARY = memoryview
ROWID = int



from typing import Any

from braintree.configuration import Configuration as Configuration
from braintree.resource import Resource as Resource

class PartnerMerchant(Resource):
    partner_merchant_id: Any
    private_key: Any
    public_key: Any
    merchant_public_id: Any
    client_side_encryption_key: Any
    def __init__(self, gateway, attributes) -> None: ...



from typing import Any, Match

class JSONDecoder(object):
    def __init__(self, **kwargs: Any) -> None: ...
    def decode(self, s: str, _w: Match[str], _PY3: bool): ...
    def raw_decode(self, s: str, idx: int, _w: Match[str], _PY3: bool): ...



from typing import Mapping, Sequence

_Cap = dict[str, str | int]

def findmatch(
    caps: Mapping[str, list[_Cap]], MIMEtype: str, key: str = ..., filename: str = ..., plist: Sequence[str] = ...
) -> tuple[str | None, _Cap | None]: ...
def getcaps() -> dict[str, list[_Cap]]: ...



from typing import Any

from openpyxl.descriptors.serialisable import Serialisable

def collapse_cell_addresses(cells, input_ranges=...): ...
def expand_cell_ranges(range_string): ...

class DataValidation(Serialisable):
    tagname: str
    sqref: Any
    cells: Any
    ranges: Any
    showErrorMessage: Any
    showDropDown: Any
    hide_drop_down: Any
    showInputMessage: Any
    allowBlank: Any
    allow_blank: Any
    errorTitle: Any
    error: Any
    promptTitle: Any
    prompt: Any
    formula1: Any
    formula2: Any
    type: Any
    errorStyle: Any
    imeMode: Any
    operator: Any
    validation_type: Any
    def __init__(
        self,
        type: Any | None = ...,
        formula1: Any | None = ...,
        formula2: Any | None = ...,
        showErrorMessage: bool = ...,
        showInputMessage: bool = ...,
        showDropDown: Any | None = ...,
        allowBlank: Any | None = ...,
        sqref=...,
        promptTitle: Any | None = ...,
        errorStyle: Any | None = ...,
        error: Any | None = ...,
        prompt: Any | None = ...,
        errorTitle: Any | None = ...,
        imeMode: Any | None = ...,
        operator: Any | None = ...,
        allow_blank: Any | None = ...,
    ) -> None: ...
    def add(self, cell) -> None: ...
    def __contains__(self, cell): ...

class DataValidationList(Serialisable):
    tagname: str
    disablePrompts: Any
    xWindow: Any
    yWindow: Any
    dataValidation: Any
    __elements__: Any
    __attrs__: Any
    def __init__(
        self,
        disablePrompts: Any | None = ...,
        xWindow: Any | None = ...,
        yWindow: Any | None = ...,
        count: Any | None = ...,
        dataValidation=...,
    ) -> None: ...
    @property
    def count(self): ...
    def __len__(self): ...
    def append(self, dv) -> None: ...
    def to_tree(self, tagname: Any | None = ...): ...  # type: ignore[override]



"""Functions and visitors for transforming pytd."""

from pytype.pytd import optimize
from pytype.pytd import visitors


def RemoveMutableParameters(ast):
  """Change all mutable parameters in a pytd AST to a non-mutable form."""
  ast = ast.Visit(optimize.AbsorbMutableParameters())
  ast = ast.Visit(optimize.CombineContainers())
  ast = ast.Visit(optimize.MergeTypeParameters())
  ast = ast.Visit(visitors.AdjustSelf(force=True))
  return ast



from _typeshed import FileDescriptor
from typing import (
    IO,
    Any,
    Callable,
    Generator,
    ItemsView,
    Iterable,
    Iterator,
    KeysView,
    MutableSequence,
    Sequence,
    Text,
    TypeVar,
    overload,
)

VERSION: str

class ParseError(SyntaxError):
    code: int
    position: tuple[int, int]

def iselement(element: object) -> bool: ...

_T = TypeVar("_T")

# Type for parser inputs. Parser will accept any unicode/str/bytes and coerce,
# and this is true in py2 and py3 (even fromstringlist() in python3 can be
# called with a heterogeneous list)
_parser_input_type = bytes | Text

# Type for individual tag/attr/ns/text values in args to most functions.
# In py2, the library accepts str or unicode everywhere and coerces
# aggressively.
# In py3, bytes is not coerced to str and so use of bytes is probably an error,
# so we exclude it. (why? the parser never produces bytes when it parses XML,
# so e.g., element.get(b'name') will always return None for parsed XML, even if
# there is a 'name' attribute.)
_str_argument_type = str | Text

# Type for return values from individual tag/attr/text values
# in python2, if the tag/attribute/text wasn't decode-able as ascii, it
# comes out as a unicode string; otherwise it comes out as str. (see
# _fixtext function in the source). Client code knows best:
_str_result_type = Any

_file_or_filename = Text | FileDescriptor | IO[Any]

class Element(MutableSequence[Element]):
    tag: _str_result_type
    attrib: dict[_str_result_type, _str_result_type]
    text: _str_result_type | None
    tail: _str_result_type | None
    def __init__(
        self,
        tag: _str_argument_type | Callable[..., Element],
        attrib: dict[_str_argument_type, _str_argument_type] = ...,
        **extra: _str_argument_type,
    ) -> None: ...
    def append(self, __subelement: Element) -> None: ...
    def clear(self) -> None: ...
    def extend(self, __elements: Iterable[Element]) -> None: ...
    def find(
        self, path: _str_argument_type, namespaces: dict[_str_argument_type, _str_argument_type] | None = ...
    ) -> Element | None: ...
    def findall(
        self, path: _str_argument_type, namespaces: dict[_str_argument_type, _str_argument_type] | None = ...
    ) -> list[Element]: ...
    @overload
    def findtext(
        self, path: _str_argument_type, default: None = ..., namespaces: dict[_str_argument_type, _str_argument_type] | None = ...
    ) -> _str_result_type | None: ...
    @overload
    def findtext(
        self, path: _str_argument_type, default: _T, namespaces: dict[_str_argument_type, _str_argument_type] | None = ...
    ) -> _T | _str_result_type: ...
    @overload
    def get(self, key: _str_argument_type, default: None = ...) -> _str_result_type | None: ...
    @overload
    def get(self, key: _str_argument_type, default: _T) -> _str_result_type | _T: ...
    def insert(self, __index: int, __element: Element) -> None: ...
    def items(self) -> ItemsView[_str_result_type, _str_result_type]: ...
    def iter(self, tag: _str_argument_type | None = ...) -> Generator[Element, None, None]: ...
    def iterfind(
        self, path: _str_argument_type, namespaces: dict[_str_argument_type, _str_argument_type] | None = ...
    ) -> Generator[Element, None, None]: ...
    def itertext(self) -> Generator[_str_result_type, None, None]: ...
    def keys(self) -> KeysView[_str_result_type]: ...
    def makeelement(self, __tag: _str_argument_type, __attrib: dict[_str_argument_type, _str_argument_type]) -> Element: ...
    def remove(self, __subelement: Element) -> None: ...
    def set(self, __key: _str_argument_type, __value: _str_argument_type) -> None: ...
    def __delitem__(self, i: int | slice) -> None: ...
    @overload
    def __getitem__(self, i: int) -> Element: ...
    @overload
    def __getitem__(self, s: slice) -> MutableSequence[Element]: ...
    def __len__(self) -> int: ...
    @overload
    def __setitem__(self, i: int, o: Element) -> None: ...
    @overload
    def __setitem__(self, s: slice, o: Iterable[Element]) -> None: ...
    def getchildren(self) -> list[Element]: ...
    def getiterator(self, tag: _str_argument_type | None = ...) -> list[Element]: ...

def SubElement(
    parent: Element,
    tag: _str_argument_type,
    attrib: dict[_str_argument_type, _str_argument_type] = ...,
    **extra: _str_argument_type,
) -> Element: ...
def Comment(text: _str_argument_type | None = ...) -> Element: ...
def ProcessingInstruction(target: _str_argument_type, text: _str_argument_type | None = ...) -> Element: ...

PI: Callable[..., Element]

class QName:
    text: str
    def __init__(self, text_or_uri: _str_argument_type, tag: _str_argument_type | None = ...) -> None: ...

class ElementTree:
    def __init__(self, element: Element | None = ..., file: _file_or_filename | None = ...) -> None: ...
    def getroot(self) -> Element: ...
    def parse(self, source: _file_or_filename, parser: XMLParser | None = ...) -> Element: ...
    def iter(self, tag: _str_argument_type | None = ...) -> Generator[Element, None, None]: ...
    def getiterator(self, tag: _str_argument_type | None = ...) -> list[Element]: ...
    def find(
        self, path: _str_argument_type, namespaces: dict[_str_argument_type, _str_argument_type] | None = ...
    ) -> Element | None: ...
    @overload
    def findtext(
        self, path: _str_argument_type, default: None = ..., namespaces: dict[_str_argument_type, _str_argument_type] | None = ...
    ) -> _str_result_type | None: ...
    @overload
    def findtext(
        self, path: _str_argument_type, default: _T, namespaces: dict[_str_argument_type, _str_argument_type] | None = ...
    ) -> _T | _str_result_type: ...
    def findall(
        self, path: _str_argument_type, namespaces: dict[_str_argument_type, _str_argument_type] | None = ...
    ) -> list[Element]: ...
    def iterfind(
        self, path: _str_argument_type, namespaces: dict[_str_argument_type, _str_argument_type] | None = ...
    ) -> Generator[Element, None, None]: ...
    def write(
        self,
        file_or_filename: _file_or_filename,
        encoding: str | None = ...,
        xml_declaration: bool | None = ...,
        default_namespace: _str_argument_type | None = ...,
        method: str | None = ...,
    ) -> None: ...
    def write_c14n(self, file: _file_or_filename) -> None: ...

def register_namespace(prefix: _str_argument_type, uri: _str_argument_type) -> None: ...
def tostring(element: Element, encoding: str | None = ..., method: str | None = ...) -> bytes: ...
def tostringlist(element: Element, encoding: str | None = ..., method: str | None = ...) -> list[bytes]: ...
def dump(elem: Element) -> None: ...
def parse(source: _file_or_filename, parser: XMLParser | None = ...) -> ElementTree: ...
def iterparse(
    source: _file_or_filename, events: Sequence[str] | None = ..., parser: XMLParser | None = ...
) -> Iterator[tuple[str, Any]]: ...
def XML(text: _parser_input_type, parser: XMLParser | None = ...) -> Element: ...
def XMLID(text: _parser_input_type, parser: XMLParser | None = ...) -> tuple[Element, dict[_str_result_type, Element]]: ...

# This is aliased to XML in the source.
fromstring = XML

def fromstringlist(sequence: Sequence[_parser_input_type], parser: XMLParser | None = ...) -> Element: ...

# This type is both not precise enough and too precise. The TreeBuilder
# requires the elementfactory to accept tag and attrs in its args and produce
# some kind of object that has .text and .tail properties.
# I've chosen to constrain the ElementFactory to always produce an Element
# because that is how almost everyone will use it.
# Unfortunately, the type of the factory arguments is dependent on how
# TreeBuilder is called by client code (they could pass strs, bytes or whatever);
# but we don't want to use a too-broad type, or it would be too hard to write
# elementfactories.
_ElementFactory = Callable[[Any, dict[Any, Any]], Element]

class TreeBuilder:
    def __init__(self, element_factory: _ElementFactory | None = ...) -> None: ...
    def close(self) -> Element: ...
    def data(self, __data: _parser_input_type) -> None: ...
    def start(self, __tag: _parser_input_type, __attrs: dict[_parser_input_type, _parser_input_type]) -> Element: ...
    def end(self, __tag: _parser_input_type) -> Element: ...

class XMLParser:
    parser: Any
    target: Any
    # TODO-what is entity used for???
    entity: Any
    version: str
    def __init__(self, html: int = ..., target: Any = ..., encoding: str | None = ...) -> None: ...
    def doctype(self, __name: str, __pubid: str, __system: str) -> None: ...
    def close(self) -> Any: ...
    def feed(self, __data: _parser_input_type) -> None: ...



from typing import Any

from openpyxl.descriptors.serialisable import Serialisable

class Protection(Serialisable):
    tagname: str
    locked: Any
    hidden: Any
    def __init__(self, locked: bool = ..., hidden: bool = ...) -> None: ...



from _typeshed import SupportsNoArgReadline
from typing import IO, Any, Sequence

DEFAULTSECT: str
MAX_INTERPOLATION_DEPTH: int

class Error(Exception):
    message: Any
    def __init__(self, msg: str = ...) -> None: ...
    def _get_message(self) -> None: ...
    def _set_message(self, value: str) -> None: ...

class NoSectionError(Error):
    section: str
    def __init__(self, section: str) -> None: ...

class DuplicateSectionError(Error):
    section: str
    def __init__(self, section: str) -> None: ...

class NoOptionError(Error):
    section: str
    option: str
    def __init__(self, option: str, section: str) -> None: ...

class InterpolationError(Error):
    section: str
    option: str
    msg: str
    def __init__(self, option: str, section: str, msg: str) -> None: ...

class InterpolationMissingOptionError(InterpolationError):
    reference: str
    def __init__(self, option: str, section: str, rawval: str, reference: str) -> None: ...

class InterpolationSyntaxError(InterpolationError): ...

class InterpolationDepthError(InterpolationError):
    def __init__(self, option: str, section: str, rawval: str) -> None: ...

class ParsingError(Error):
    filename: str
    errors: list[tuple[Any, Any]]
    def __init__(self, filename: str) -> None: ...
    def append(self, lineno: Any, line: Any) -> None: ...

class MissingSectionHeaderError(ParsingError):
    lineno: Any
    line: Any
    def __init__(self, filename: str, lineno: Any, line: Any) -> None: ...

class RawConfigParser:
    _dict: Any
    _sections: dict[Any, Any]
    _defaults: dict[Any, Any]
    _optcre: Any
    SECTCRE: Any
    OPTCRE: Any
    OPTCRE_NV: Any
    def __init__(self, defaults: dict[Any, Any] = ..., dict_type: Any = ..., allow_no_value: bool = ...) -> None: ...
    def defaults(self) -> dict[Any, Any]: ...
    def sections(self) -> list[str]: ...
    def add_section(self, section: str) -> None: ...
    def has_section(self, section: str) -> bool: ...
    def options(self, section: str) -> list[str]: ...
    def read(self, filenames: str | Sequence[str]) -> list[str]: ...
    def readfp(self, fp: SupportsNoArgReadline[str], filename: str = ...) -> None: ...
    def get(self, section: str, option: str) -> str: ...
    def items(self, section: str) -> list[tuple[Any, Any]]: ...
    def _get(self, section: str, conv: type, option: str) -> Any: ...
    def getint(self, section: str, option: str) -> int: ...
    def getfloat(self, section: str, option: str) -> float: ...
    _boolean_states: dict[str, bool]
    def getboolean(self, section: str, option: str) -> bool: ...
    def optionxform(self, optionstr: str) -> str: ...
    def has_option(self, section: str, option: str) -> bool: ...
    def set(self, section: str, option: str, value: Any = ...) -> None: ...
    def write(self, fp: IO[str]) -> None: ...
    def remove_option(self, section: str, option: Any) -> bool: ...
    def remove_section(self, section: str) -> bool: ...
    def _read(self, fp: IO[str], fpname: str) -> None: ...

class ConfigParser(RawConfigParser):
    _KEYCRE: Any
    def get(self, section: str, option: str, raw: bool = ..., vars: dict[Any, Any] | None = ...) -> Any: ...
    def items(self, section: str, raw: bool = ..., vars: dict[Any, Any] | None = ...) -> list[tuple[str, Any]]: ...
    def _interpolate(self, section: str, option: str, rawval: Any, vars: Any) -> str: ...
    def _interpolation_replace(self, match: Any) -> str: ...

class SafeConfigParser(ConfigParser):
    _interpvar_re: Any
    def _interpolate(self, section: str, option: str, rawval: Any, vars: Any) -> str: ...
    def _interpolate_some(
        self, option: str, accum: list[Any], rest: str, section: str, map: dict[Any, Any], depth: int
    ) -> None: ...



from distutils.errors import DistutilsError
from typing import Any

class UnrecognizedFormat(DistutilsError): ...

def default_filter(src, dst): ...
def unpack_archive(filename, extract_dir, progress_filter=..., drivers: Any | None = ...) -> None: ...
def unpack_directory(filename, extract_dir, progress_filter=...) -> None: ...
def unpack_zipfile(filename, extract_dir, progress_filter=...) -> None: ...
def unpack_tarfile(filename, extract_dir, progress_filter=...): ...

extraction_drivers: Any



"""Tests for compare.py."""

from pytype import compare
from pytype import config
from pytype.abstract import abstract
from pytype.abstract import abstract_utils
from pytype.abstract import function
from pytype.pytd import pytd
from pytype.pytd import slots
from pytype.tests import test_base
from pytype.tests import test_utils

import unittest


class CompareTestBase(test_base.UnitTest):

  def setUp(self):
    super().setUp()
    options = config.Options.create(python_version=self.python_version)
    self._ctx = test_utils.make_context(options)
    self._program = self._ctx.program
    self._node = self._ctx.root_node.ConnectNew("test_node")
    self._convert = self._ctx.convert

  def assertTruthy(self, value):
    self.assertIs(True, compare.compatible_with(value, True))
    self.assertIs(False, compare.compatible_with(value, False))

  def assertFalsy(self, value):
    self.assertIs(False, compare.compatible_with(value, True))
    self.assertIs(True, compare.compatible_with(value, False))

  def assertAmbiguous(self, value):
    self.assertIs(True, compare.compatible_with(value, True))
    self.assertIs(True, compare.compatible_with(value, False))


class InstanceTest(CompareTestBase):

  def test_compatible_with_object(self):
    # object() is not compatible with False
    i = abstract.Instance(self._convert.object_type, self._ctx)
    self.assertTruthy(i)

  def test_compatible_with_numeric(self):
    # Numbers can evaluate to True or False
    i = abstract.Instance(self._convert.int_type, self._ctx)
    self.assertAmbiguous(i)

  def test_compatible_with_list(self):
    i = abstract.List([], self._ctx)
    # Empty list is not compatible with True.
    self.assertFalsy(i)
    # Once a type parameter is set, list is compatible with True and False.
    i.merge_instance_type_parameter(
        self._node, abstract_utils.T,
        self._convert.object_type.to_variable(self._ctx.root_node))
    self.assertAmbiguous(i)

  def test_compatible_with_set(self):
    i = abstract.Instance(self._convert.set_type, self._ctx)
    # Empty set is not compatible with True.
    self.assertFalsy(i)
    # Once a type parameter is set, list is compatible with True and False.
    i.merge_instance_type_parameter(
        self._node, abstract_utils.T,
        self._convert.object_type.to_variable(self._ctx.root_node))
    self.assertAmbiguous(i)

  def test_compatible_with_none(self):
    # This test is specifically for abstract.Instance, so we don't use
    # self._convert.none, which is a ConcreteValue.
    i = abstract.Instance(self._convert.none_type, self._ctx)
    self.assertFalsy(i)

  def test_compare_frozensets(self):
    """Test that two frozensets can be compared for equality."""
    fset = self._convert.frozenset_type
    i = abstract.Instance(fset, self._ctx)
    j = abstract.Instance(fset, self._ctx)
    self.assertIs(None, compare.cmp_rel(self._ctx, slots.EQ, i, j))


class TupleTest(CompareTestBase):

  def setUp(self):
    super().setUp()
    self._var = self._program.NewVariable()
    self._var.AddBinding(abstract.Unknown(self._ctx), [], self._node)

  def test_compatible_with__not_empty(self):
    t = self._ctx.convert.tuple_to_value((self._var,))
    self.assertTruthy(t)

  def test_compatible_with__empty(self):
    t = self._ctx.convert.tuple_to_value(())
    self.assertFalsy(t)

  def test_getitem__concrete_index(self):
    t = self._ctx.convert.tuple_to_value((self._var,))
    index = self._convert.constant_to_var(0)
    node, var = t.cls.getitem_slot(self._node, index)
    self.assertIs(node, self._node)
    self.assertIs(abstract_utils.get_atomic_value(var),
                  abstract_utils.get_atomic_value(self._var))

  def test_getitem__abstract_index(self):
    t = self._ctx.convert.tuple_to_value((self._var,))
    index = self._convert.build_int(self._node)
    node, var = t.cls.getitem_slot(self._node, index)
    self.assertIs(node, self._node)
    self.assertIs(abstract_utils.get_atomic_value(var),
                  abstract_utils.get_atomic_value(self._var))

  def test_cmp_rel__equal(self):
    tup = self._convert.constant_to_value((3, 1))
    self.assertIs(False, compare.cmp_rel(self._ctx, slots.LT, tup, tup))
    self.assertIs(True, compare.cmp_rel(self._ctx, slots.LE, tup, tup))
    self.assertIs(True, compare.cmp_rel(self._ctx, slots.EQ, tup, tup))
    self.assertIs(False, compare.cmp_rel(self._ctx, slots.NE, tup, tup))
    self.assertIs(True, compare.cmp_rel(self._ctx, slots.GE, tup, tup))
    self.assertIs(False, compare.cmp_rel(self._ctx, slots.GT, tup, tup))

  def test_cmp_rel__not_equal(self):
    tup1 = self._convert.constant_to_value((3, 1))
    tup2 = self._convert.constant_to_value((3, 5))
    self.assertIs(True, compare.cmp_rel(self._ctx, slots.LT, tup1, tup2))
    self.assertIs(False, compare.cmp_rel(self._ctx, slots.LT, tup2, tup1))
    self.assertIs(True, compare.cmp_rel(self._ctx, slots.LE, tup1, tup2))
    self.assertIs(False, compare.cmp_rel(self._ctx, slots.LE, tup2, tup1))
    self.assertIs(False, compare.cmp_rel(self._ctx, slots.EQ, tup1, tup2))
    self.assertIs(False, compare.cmp_rel(self._ctx, slots.EQ, tup2, tup1))
    self.assertIs(True, compare.cmp_rel(self._ctx, slots.NE, tup1, tup2))
    self.assertIs(True, compare.cmp_rel(self._ctx, slots.NE, tup2, tup1))
    self.assertIs(False, compare.cmp_rel(self._ctx, slots.GE, tup1, tup2))
    self.assertIs(True, compare.cmp_rel(self._ctx, slots.GE, tup2, tup1))
    self.assertIs(False, compare.cmp_rel(self._ctx, slots.GT, tup1, tup2))
    self.assertIs(True, compare.cmp_rel(self._ctx, slots.GT, tup2, tup1))

  def test_cmp_rel__unknown(self):
    tup1 = self._convert.constant_to_value((3, 1))
    tup2 = abstract.Instance(self._convert.tuple_type, self._ctx)
    for op in (slots.LT, slots.LE, slots.EQ, slots.NE, slots.GE, slots.GT):
      self.assertIsNone(compare.cmp_rel(self._ctx, op, tup1, tup2))
      self.assertIsNone(compare.cmp_rel(self._ctx, op, tup2, tup1))

  def test_cmp_rel__prefix_equal(self):
    tup1 = self._ctx.convert.tuple_to_value(
        (self._convert.constant_to_value(3).to_variable(self._node),
         self._convert.constant_to_value(1).to_variable(self._node),
         self._convert.primitive_class_instances[int].to_variable(self._node)))
    tup2 = self._convert.constant_to_value((3, 1))
    self.assertIs(False, compare.cmp_rel(self._ctx, slots.LT, tup1, tup2))
    self.assertIs(True, compare.cmp_rel(self._ctx, slots.LT, tup2, tup1))
    self.assertIs(False, compare.cmp_rel(self._ctx, slots.LE, tup1, tup2))
    self.assertIs(True, compare.cmp_rel(self._ctx, slots.LE, tup2, tup1))
    self.assertIs(False, compare.cmp_rel(self._ctx, slots.EQ, tup1, tup2))
    self.assertIs(False, compare.cmp_rel(self._ctx, slots.EQ, tup2, tup1))
    self.assertIs(True, compare.cmp_rel(self._ctx, slots.NE, tup1, tup2))
    self.assertIs(True, compare.cmp_rel(self._ctx, slots.NE, tup2, tup1))
    self.assertIs(True, compare.cmp_rel(self._ctx, slots.GE, tup1, tup2))
    self.assertIs(False, compare.cmp_rel(self._ctx, slots.GE, tup2, tup1))
    self.assertIs(True, compare.cmp_rel(self._ctx, slots.GT, tup1, tup2))
    self.assertIs(False, compare.cmp_rel(self._ctx, slots.GT, tup2, tup1))

  def test_cmp_rel__prefix_not_equal(self):
    tup1 = self._ctx.convert.tuple_to_value(
        (self._convert.constant_to_value(3).to_variable(self._node),
         self._convert.constant_to_value(1).to_variable(self._node),
         self._convert.primitive_class_instances[int].to_variable(self._node)))
    tup2 = self._convert.constant_to_value((4, 2))
    self.assertIs(True, compare.cmp_rel(self._ctx, slots.LT, tup1, tup2))
    self.assertIs(False, compare.cmp_rel(self._ctx, slots.LT, tup2, tup1))
    self.assertIs(True, compare.cmp_rel(self._ctx, slots.LE, tup1, tup2))
    self.assertIs(False, compare.cmp_rel(self._ctx, slots.LE, tup2, tup1))
    self.assertIs(False, compare.cmp_rel(self._ctx, slots.EQ, tup1, tup2))
    self.assertIs(False, compare.cmp_rel(self._ctx, slots.EQ, tup2, tup1))
    self.assertIs(True, compare.cmp_rel(self._ctx, slots.NE, tup1, tup2))
    self.assertIs(True, compare.cmp_rel(self._ctx, slots.NE, tup2, tup1))
    self.assertIs(False, compare.cmp_rel(self._ctx, slots.GE, tup1, tup2))
    self.assertIs(True, compare.cmp_rel(self._ctx, slots.GE, tup2, tup1))
    self.assertIs(False, compare.cmp_rel(self._ctx, slots.GT, tup1, tup2))
    self.assertIs(True, compare.cmp_rel(self._ctx, slots.GT, tup2, tup1))

  def test_cmp_rel__prefix_unknown(self):
    tup1 = self._ctx.convert.tuple_to_value(
        (self._convert.constant_to_value(3).to_variable(self._node),
         self._convert.primitive_class_instances[int].to_variable(self._node)))
    tup2 = self._convert.constant_to_value((3, 1))
    for op in (slots.LT, slots.LE, slots.EQ, slots.NE, slots.GE, slots.GT):
      self.assertIsNone(compare.cmp_rel(self._ctx, op, tup1, tup2))
      self.assertIsNone(compare.cmp_rel(self._ctx, op, tup2, tup1))


class DictTest(CompareTestBase):

  def setUp(self):
    super().setUp()
    self._d = abstract.Dict(self._ctx)
    self._var = self._program.NewVariable()
    self._var.AddBinding(abstract.Unknown(self._ctx), [], self._node)

  def test_compatible_with__when_empty(self):
    self.assertFalsy(self._d)

  def test_compatible_with__after_setitem(self):
    # Once a slot is added, dict is ambiguous.
    self._d.setitem_slot(self._node, self._var, self._var)
    self.assertAmbiguous(self._d)

  def test_compatible_with__after_set_str_item(self):
    self._d.set_str_item(self._node, "key", self._var)
    self.assertTruthy(self._d)

  def test_compatible_with__after_unknown_update(self):
    # Updating an empty dict with an unknown value makes the former ambiguous.
    self._d.update(self._node, abstract.Unknown(self._ctx))
    self.assertAmbiguous(self._d)

  def test_compatible_with__after_empty_update(self):
    empty_dict = abstract.Dict(self._ctx)
    self._d.update(self._node, empty_dict)
    self.assertFalsy(self._d)

  def test_compatible_with__after_unambiguous_update(self):
    unambiguous_dict = abstract.Dict(self._ctx)
    unambiguous_dict.set_str_item(self._node, "a",
                                  self._ctx.new_unsolvable(self._node))
    self._d.update(self._node, unambiguous_dict)
    self.assertTruthy(self._d)

  def test_compatible_with__after_ambiguous_update(self):
    ambiguous_dict = abstract.Dict(self._ctx)
    ambiguous_dict.merge_instance_type_parameter(
        self._node, abstract_utils.K, self._ctx.new_unsolvable(self._node))
    ambiguous_dict.could_contain_anything = True
    self._d.update(self._node, ambiguous_dict)
    self.assertAmbiguous(self._d)

  def test_compatible_with__after_concrete_update(self):
    self._d.update(self._node, {})
    self.assertFalsy(self._d)
    self._d.update(self._node, {"a": self._ctx.new_unsolvable(self._node)})
    self.assertTruthy(self._d)

  def test_pop(self):
    self._d.set_str_item(self._node, "a", self._var)
    node, ret = self._d.pop_slot(self._node,
                                 self._convert.build_string(self._node, "a"))
    self.assertFalsy(self._d)
    self.assertIs(node, self._node)
    self.assertIs(ret, self._var)

  def test_pop_with_default(self):
    self._d.set_str_item(self._node, "a", self._var)
    node, ret = self._d.pop_slot(self._node,
                                 self._convert.build_string(self._node, "a"),
                                 self._convert.none.to_variable(
                                     self._node))  # default is ignored
    self.assertFalsy(self._d)
    self.assertIs(node, self._node)
    self.assertIs(ret, self._var)

  def test_bad_pop(self):
    self._d.set_str_item(self._node, "a", self._var)
    self.assertRaises(function.DictKeyMissing, self._d.pop_slot, self._node,
                      self._convert.build_string(self._node, "b"))
    self.assertTruthy(self._d)

  def test_bad_pop_with_default(self):
    val = self._convert.primitive_class_instances[int]
    self._d.set_str_item(self._node, "a", val.to_variable(self._node))
    node, ret = self._d.pop_slot(self._node,
                                 self._convert.build_string(self._node, "b"),
                                 self._convert.none.to_variable(self._node))
    self.assertTruthy(self._d)
    self.assertIs(node, self._node)
    self.assertListEqual(ret.data, [self._convert.none])

  def test_ambiguous_pop(self):
    val = self._convert.primitive_class_instances[int]
    self._d.set_str_item(self._node, "a", val.to_variable(self._node))
    ambiguous_key = self._convert.primitive_class_instances[str]
    node, ret = self._d.pop_slot(
        self._node, ambiguous_key.to_variable(self._node))
    self.assertAmbiguous(self._d)
    self.assertIs(node, self._node)
    self.assertListEqual(ret.data, [val])

  def test_ambiguous_pop_with_default(self):
    val = self._convert.primitive_class_instances[int]
    self._d.set_str_item(self._node, "a", val.to_variable(self._node))
    ambiguous_key = self._convert.primitive_class_instances[str]
    default_var = self._convert.none.to_variable(self._node)
    node, ret = self._d.pop_slot(
        self._node, ambiguous_key.to_variable(self._node), default_var)
    self.assertAmbiguous(self._d)
    self.assertIs(node, self._node)
    self.assertSetEqual(set(ret.data), {val, self._convert.none})

  def test_ambiguous_dict_after_pop(self):
    ambiguous_key = self._convert.primitive_class_instances[str]
    val = self._convert.primitive_class_instances[int]
    node, _ = self._d.setitem_slot(
        self._node, ambiguous_key.to_variable(self._node),
        val.to_variable(self._node))
    _, ret = self._d.pop_slot(node, self._convert.build_string(node, "a"))
    self.assertAmbiguous(self._d)
    self.assertListEqual(ret.data, [val])

  def test_ambiguous_dict_after_pop_with_default(self):
    ambiguous_key = self._convert.primitive_class_instances[str]
    val = self._convert.primitive_class_instances[int]
    node, _ = self._d.setitem_slot(
        self._node, ambiguous_key.to_variable(self._node),
        val.to_variable(self._node))
    _, ret = self._d.pop_slot(node, self._convert.build_string(node, "a"),
                              self._convert.none.to_variable(node))
    self.assertAmbiguous(self._d)
    self.assertSetEqual(set(ret.data), {val, self._convert.none})


class FunctionTest(CompareTestBase):

  def test_compatible_with(self):
    pytd_sig = pytd.Signature((), None, None, pytd.AnythingType(), (), ())
    sig = abstract.PyTDSignature("f", pytd_sig, self._ctx)
    f = abstract.PyTDFunction("f", (sig,), pytd.MethodKind.METHOD, self._ctx)
    self.assertTruthy(f)


class ClassTest(CompareTestBase):

  def test_compatible_with(self):
    cls = abstract.InterpreterClass("X", [], {}, None, self._ctx)
    self.assertTruthy(cls)


if __name__ == "__main__":
  unittest.main()



import sys
from lib2to3.pgen2.token import *
from typing import Callable, Iterable, Iterator

if sys.version_info >= (3, 8):
    __all__ = [
        "AMPER",
        "AMPEREQUAL",
        "ASYNC",
        "AT",
        "ATEQUAL",
        "AWAIT",
        "BACKQUOTE",
        "CIRCUMFLEX",
        "CIRCUMFLEXEQUAL",
        "COLON",
        "COLONEQUAL",
        "COMMA",
        "COMMENT",
        "DEDENT",
        "DOT",
        "DOUBLESLASH",
        "DOUBLESLASHEQUAL",
        "DOUBLESTAR",
        "DOUBLESTAREQUAL",
        "ENDMARKER",
        "EQEQUAL",
        "EQUAL",
        "ERRORTOKEN",
        "GREATER",
        "GREATEREQUAL",
        "INDENT",
        "ISEOF",
        "ISNONTERMINAL",
        "ISTERMINAL",
        "LBRACE",
        "LEFTSHIFT",
        "LEFTSHIFTEQUAL",
        "LESS",
        "LESSEQUAL",
        "LPAR",
        "LSQB",
        "MINEQUAL",
        "MINUS",
        "NAME",
        "NEWLINE",
        "NL",
        "NOTEQUAL",
        "NT_OFFSET",
        "NUMBER",
        "N_TOKENS",
        "OP",
        "PERCENT",
        "PERCENTEQUAL",
        "PLUS",
        "PLUSEQUAL",
        "RARROW",
        "RBRACE",
        "RIGHTSHIFT",
        "RIGHTSHIFTEQUAL",
        "RPAR",
        "RSQB",
        "SEMI",
        "SLASH",
        "SLASHEQUAL",
        "STAR",
        "STAREQUAL",
        "STRING",
        "TILDE",
        "VBAR",
        "VBAREQUAL",
        "tok_name",
        "tokenize",
        "generate_tokens",
        "untokenize",
    ]
else:
    __all__ = [
        "AMPER",
        "AMPEREQUAL",
        "ASYNC",
        "AT",
        "ATEQUAL",
        "AWAIT",
        "BACKQUOTE",
        "CIRCUMFLEX",
        "CIRCUMFLEXEQUAL",
        "COLON",
        "COMMA",
        "COMMENT",
        "DEDENT",
        "DOT",
        "DOUBLESLASH",
        "DOUBLESLASHEQUAL",
        "DOUBLESTAR",
        "DOUBLESTAREQUAL",
        "ENDMARKER",
        "EQEQUAL",
        "EQUAL",
        "ERRORTOKEN",
        "GREATER",
        "GREATEREQUAL",
        "INDENT",
        "ISEOF",
        "ISNONTERMINAL",
        "ISTERMINAL",
        "LBRACE",
        "LEFTSHIFT",
        "LEFTSHIFTEQUAL",
        "LESS",
        "LESSEQUAL",
        "LPAR",
        "LSQB",
        "MINEQUAL",
        "MINUS",
        "NAME",
        "NEWLINE",
        "NL",
        "NOTEQUAL",
        "NT_OFFSET",
        "NUMBER",
        "N_TOKENS",
        "OP",
        "PERCENT",
        "PERCENTEQUAL",
        "PLUS",
        "PLUSEQUAL",
        "RARROW",
        "RBRACE",
        "RIGHTSHIFT",
        "RIGHTSHIFTEQUAL",
        "RPAR",
        "RSQB",
        "SEMI",
        "SLASH",
        "SLASHEQUAL",
        "STAR",
        "STAREQUAL",
        "STRING",
        "TILDE",
        "VBAR",
        "VBAREQUAL",
        "tok_name",
        "tokenize",
        "generate_tokens",
        "untokenize",
    ]

_Coord = tuple[int, int]
_TokenEater = Callable[[int, str, _Coord, _Coord, str], None]
_TokenInfo = tuple[int, str, _Coord, _Coord, str]

class TokenError(Exception): ...
class StopTokenizing(Exception): ...

def tokenize(readline: Callable[[], str], tokeneater: _TokenEater = ...) -> None: ...

class Untokenizer:
    tokens: list[str]
    prev_row: int
    prev_col: int
    def __init__(self) -> None: ...
    def add_whitespace(self, start: _Coord) -> None: ...
    def untokenize(self, iterable: Iterable[_TokenInfo]) -> str: ...
    def compat(self, token: tuple[int, str], iterable: Iterable[_TokenInfo]) -> None: ...

def untokenize(iterable: Iterable[_TokenInfo]) -> str: ...
def generate_tokens(readline: Callable[[], str]) -> Iterator[_TokenInfo]: ...



# Utility types for typeshed
#
# See the README.md file in this directory for more information.

import array
import ctypes
import mmap
import sys
from os import PathLike
from typing import AbstractSet, Any, Awaitable, Container, Generic, Iterable, Protocol, TypeVar
from typing_extensions import Final, Literal, TypeAlias, final

_KT = TypeVar("_KT")
_KT_co = TypeVar("_KT_co", covariant=True)
_KT_contra = TypeVar("_KT_contra", contravariant=True)
_VT = TypeVar("_VT")
_VT_co = TypeVar("_VT_co", covariant=True)
_T = TypeVar("_T")
_T_co = TypeVar("_T_co", covariant=True)
_T_contra = TypeVar("_T_contra", contravariant=True)

# Use for "self" annotations:
#   def __enter__(self: Self) -> Self: ...
Self = TypeVar("Self")  # noqa: Y001

# For partially known annotations. Usually, fields where type annotations
# haven't been added are left unannotated, but in some situations this
# isn't possible or a type is already partially known. In cases like these,
# use Incomplete instead of Any as a marker. For example, use
# "Incomplete | None" instead of "Any | None".
Incomplete: TypeAlias = Any

# stable
class IdentityFunction(Protocol):
    def __call__(self, __x: _T) -> _T: ...

# stable
class SupportsNext(Protocol[_T_co]):
    def __next__(self) -> _T_co: ...

# stable
class SupportsAnext(Protocol[_T_co]):
    def __anext__(self) -> Awaitable[_T_co]: ...

# Comparison protocols

class SupportsDunderLT(Protocol):
    def __lt__(self, __other: Any) -> bool: ...

class SupportsDunderGT(Protocol):
    def __gt__(self, __other: Any) -> bool: ...

class SupportsDunderLE(Protocol):
    def __le__(self, __other: Any) -> bool: ...

class SupportsDunderGE(Protocol):
    def __ge__(self, __other: Any) -> bool: ...

class SupportsAllComparisons(SupportsDunderLT, SupportsDunderGT, SupportsDunderLE, SupportsDunderGE, Protocol): ...

SupportsRichComparison: TypeAlias = SupportsDunderLT | SupportsDunderGT
SupportsRichComparisonT = TypeVar("SupportsRichComparisonT", bound=SupportsRichComparison)  # noqa: Y001

class SupportsDivMod(Protocol[_T_contra, _T_co]):
    def __divmod__(self, __other: _T_contra) -> _T_co: ...

class SupportsRDivMod(Protocol[_T_contra, _T_co]):
    def __rdivmod__(self, __other: _T_contra) -> _T_co: ...

class SupportsLenAndGetItem(Protocol[_T_co]):
    def __len__(self) -> int: ...
    def __getitem__(self, __k: int) -> _T_co: ...

class SupportsTrunc(Protocol):
    def __trunc__(self) -> int: ...

# Mapping-like protocols

# stable
class SupportsItems(Protocol[_KT_co, _VT_co]):
    def items(self) -> AbstractSet[tuple[_KT_co, _VT_co]]: ...

# stable
class SupportsKeysAndGetItem(Protocol[_KT, _VT_co]):
    def keys(self) -> Iterable[_KT]: ...
    def __getitem__(self, __k: _KT) -> _VT_co: ...

# stable
class SupportsGetItem(Container[_KT_contra], Protocol[_KT_contra, _VT_co]):
    def __getitem__(self, __k: _KT_contra) -> _VT_co: ...

# stable
class SupportsItemAccess(SupportsGetItem[_KT_contra, _VT], Protocol[_KT_contra, _VT]):
    def __setitem__(self, __k: _KT_contra, __v: _VT) -> None: ...
    def __delitem__(self, __v: _KT_contra) -> None: ...

# These aliases are simple strings in Python 2.
StrPath: TypeAlias = str | PathLike[str]  # stable
BytesPath: TypeAlias = bytes | PathLike[bytes]  # stable
StrOrBytesPath: TypeAlias = str | bytes | PathLike[str] | PathLike[bytes]  # stable

OpenTextModeUpdating = Literal[
    "r+",
    "+r",
    "rt+",
    "r+t",
    "+rt",
    "tr+",
    "t+r",
    "+tr",
    "w+",
    "+w",
    "wt+",
    "w+t",
    "+wt",
    "tw+",
    "t+w",
    "+tw",
    "a+",
    "+a",
    "at+",
    "a+t",
    "+at",
    "ta+",
    "t+a",
    "+ta",
    "x+",
    "+x",
    "xt+",
    "x+t",
    "+xt",
    "tx+",
    "t+x",
    "+tx",
]
OpenTextModeWriting: TypeAlias = Literal["w", "wt", "tw", "a", "at", "ta", "x", "xt", "tx"]
OpenTextModeReading: TypeAlias = Literal["r", "rt", "tr", "U", "rU", "Ur", "rtU", "rUt", "Urt", "trU", "tUr", "Utr"]
OpenTextMode: TypeAlias = OpenTextModeUpdating | OpenTextModeWriting | OpenTextModeReading
OpenBinaryModeUpdating: TypeAlias = Literal[
    "rb+",
    "r+b",
    "+rb",
    "br+",
    "b+r",
    "+br",
    "wb+",
    "w+b",
    "+wb",
    "bw+",
    "b+w",
    "+bw",
    "ab+",
    "a+b",
    "+ab",
    "ba+",
    "b+a",
    "+ba",
    "xb+",
    "x+b",
    "+xb",
    "bx+",
    "b+x",
    "+bx",
]
OpenBinaryModeWriting: TypeAlias = Literal["wb", "bw", "ab", "ba", "xb", "bx"]
OpenBinaryModeReading: TypeAlias = Literal["rb", "br", "rbU", "rUb", "Urb", "brU", "bUr", "Ubr"]
OpenBinaryMode: TypeAlias = OpenBinaryModeUpdating | OpenBinaryModeReading | OpenBinaryModeWriting

# stable
class HasFileno(Protocol):
    def fileno(self) -> int: ...

FileDescriptor: TypeAlias = int  # stable
FileDescriptorLike: TypeAlias = int | HasFileno  # stable

# stable
class SupportsRead(Protocol[_T_co]):
    def read(self, __length: int = ...) -> _T_co: ...

# stable
class SupportsReadline(Protocol[_T_co]):
    def readline(self, __length: int = ...) -> _T_co: ...

# stable
class SupportsNoArgReadline(Protocol[_T_co]):
    def readline(self) -> _T_co: ...

# stable
class SupportsWrite(Protocol[_T_contra]):
    def write(self, __s: _T_contra) -> object: ...

ReadOnlyBuffer: TypeAlias = bytes  # stable
# Anything that implements the read-write buffer interface.
# The buffer interface is defined purely on the C level, so we cannot define a normal Protocol
# for it. Instead we have to list the most common stdlib buffer classes in a Union.
WriteableBuffer: TypeAlias = bytearray | memoryview | array.array[Any] | mmap.mmap | ctypes._CData  # stable
# Same as _WriteableBuffer, but also includes read-only buffer types (like bytes).
ReadableBuffer: TypeAlias = ReadOnlyBuffer | WriteableBuffer  # stable

# stable
if sys.version_info >= (3, 10):
    from types import NoneType as NoneType
else:
    # Used by type checkers for checks involving None (does not exist at runtime)
    @final
    class NoneType:
        def __bool__(self) -> Literal[False]: ...

# This is an internal CPython type that is like, but subtly different from, a NamedTuple
# Subclasses of this type are found in multiple modules.
# In typeshed, `structseq` is only ever used as a mixin in combination with a fixed-length `Tuple`
# See discussion at #6546 & #6560
# `structseq` classes are unsubclassable, so are all decorated with `@final`.
class structseq(Generic[_T_co]):
    n_fields: Final[int]
    n_unnamed_fields: Final[int]
    n_sequence_fields: Final[int]
    # The first parameter will generally only take an iterable of a specific length.
    # E.g. `os.uname_result` takes any iterable of length exactly 5.
    #
    # The second parameter will accept a dict of any kind without raising an exception,
    # but only has any meaning if you supply it a dict where the keys are strings.
    # https://github.com/python/typeshed/pull/6560#discussion_r767149830
    def __new__(cls: type[Self], sequence: Iterable[_T_co], dict: dict[str, Any] = ...) -> Self: ...



"""Utilities for inline type annotations."""

import collections
import dataclasses
import itertools

from typing import AbstractSet, Any, Optional

from pytype import utils
from pytype.abstract import abstract
from pytype.abstract import abstract_utils
from pytype.abstract import mixin
from pytype.overlays import typing_overlay
from pytype.pytd import pytd_utils


@dataclasses.dataclass
class AnnotatedValue:
  typ: Any
  value: Any
  final: bool = False


class AnnotationUtils(utils.ContextWeakrefMixin):
  """Utility class for inline type annotations."""

  def sub_annotations(self, node, annotations, substs, instantiate_unbound):
    """Apply type parameter substitutions to a dictionary of annotations."""
    if substs and all(substs):
      return {name: self.sub_one_annotation(node, annot, substs,
                                            instantiate_unbound)
              for name, annot in annotations.items()}
    return annotations

  def _get_type_parameter_subst(self, node, annot, substs, instantiate_unbound):
    """Helper for sub_one_annotation."""
    assert isinstance(annot, abstract.TypeParameter)
    # We use the given substitutions to bind the annotation if
    # (1) every subst provides at least one binding, and
    # (2) none of the bindings are ambiguous, and
    # (3) at least one binding is non-empty.
    if all(annot.full_name in subst and subst[annot.full_name].bindings
           for subst in substs):
      vals = sum((subst[annot.full_name].data for subst in substs), [])
    else:
      vals = None
    if (vals is None or
        any(isinstance(v, abstract.AMBIGUOUS) for v in vals) or
        all(isinstance(v, abstract.Empty) for v in vals)):
      if instantiate_unbound:
        vals = annot.instantiate(node).data
      else:
        vals = [annot]
    return self.ctx.convert.merge_classes(vals)

  def sub_one_annotation(self, node, annot, substs, instantiate_unbound=True):
    """Apply type parameter substitutions to an annotation."""
    # We push annotations onto 'stack' and move them to the 'done' stack as they
    # are processed. For each annotation, we also track an 'inner_type_keys'
    # value, which is meaningful only for a NestedAnnotation. For a
    # NestedAnnotation, inner_type_keys=None indicates the annotation has not
    # yet been seen, so we push its inner types onto the stack, followed by the
    # annotation itself with its real 'inner_type_keys' value. When we see the
    # annotation again, we pull the processed inner types off the 'done' stack
    # and construct the final annotation.
    stack = [(annot, None)]
    late_annotations = {}
    done = []
    while stack:
      cur, inner_type_keys = stack.pop()
      if not cur.formal:
        done.append(cur)
      elif isinstance(cur, mixin.NestedAnnotation):
        if cur.is_late_annotation() and any(t[0] == cur for t in stack):
          # We've found a recursive type. We generate a LateAnnotation as a
          # placeholder for the substituted type.
          if cur not in late_annotations:
            param_strings = []
            for t in utils.unique_list(self.get_type_parameters(cur)):
              s = pytd_utils.Print(self._get_type_parameter_subst(
                  node, t, substs, instantiate_unbound).get_instance_type(node))
              param_strings.append(s)
            expr = f"{cur.expr}[{', '.join(param_strings)}]"
            late_annot = abstract.LateAnnotation(expr, cur.stack, cur.ctx)
            late_annotations[cur] = late_annot
          done.append(late_annotations[cur])
        elif inner_type_keys is None:
          keys, vals = zip(*cur.get_inner_types())
          stack.append((cur, keys))
          stack.extend((val, None) for val in vals)
        else:
          inner_types = []
          for k in inner_type_keys:
            inner_types.append((k, done.pop()))
          done_annot = cur.replace(inner_types)
          if cur in late_annotations:
            # If we've generated a LateAnnotation placeholder for cur's
            # substituted type, replace it now with the real type.
            late_annot = late_annotations.pop(cur)
            late_annot.set_type(done_annot)
            if "[" in late_annot.expr:
              if self.ctx.vm.late_annotations is None:
                self.ctx.vm.flatten_late_annotation(
                    node, late_annot, self.ctx.vm.frame.f_globals)
              else:
                self.ctx.vm.late_annotations[
                    late_annot.expr.split("[", 1)[0]].append(late_annot)
          done.append(done_annot)
      else:
        done.append(self._get_type_parameter_subst(
            node, cur, substs, instantiate_unbound))
    assert len(done) == 1
    return done[0]

  def get_late_annotations(self, annot):
    if annot.is_late_annotation() and not annot.resolved:
      yield annot
    elif isinstance(annot, mixin.NestedAnnotation):
      for _, typ in annot.get_inner_types():
        yield from self.get_late_annotations(typ)

  def remove_late_annotations(self, annot, seen=None):
    """Replace unresolved late annotations with unsolvables."""
    if seen is None:
      seen = {annot}
    elif annot in seen:
      return annot
    else:
      seen.add(annot)
    if annot.is_late_annotation() and not annot.resolved:
      return self.ctx.convert.unsolvable
    elif isinstance(annot, mixin.NestedAnnotation):
      inner_types = [(key, self.remove_late_annotations(val, seen))
                     for key, val in annot.get_inner_types()]
      return annot.replace(inner_types)
    return annot

  def add_scope(self, annot, types, module, seen=None):
    """Add scope for type parameters.

    In original type class, all type parameters that should be added a scope
    will be replaced with a new copy.

    Args:
      annot: The type class.
      types: A type name list that should be added a scope.
      module: Module name.
      seen: Already seen types.

    Returns:
      The type with fresh type parameters that have been added the scope.
    """
    if seen is None:
      seen = {annot}
    elif annot in seen:
      return annot
    else:
      seen.add(annot)
    if isinstance(annot, abstract.TypeParameter):
      if annot.name in types:
        new_annot = annot.copy()
        new_annot.module = module
        return new_annot
      return annot
    elif isinstance(annot, mixin.NestedAnnotation):
      inner_types = [(key, self.add_scope(typ, types, module, seen))
                     for key, typ in annot.get_inner_types()]
      return annot.replace(inner_types)
    return annot

  def get_type_parameters(self, annot, seen=None):
    """Returns all the TypeParameter instances that appear in the annotation.

    Note that if you just need to know whether or not the annotation contains
    type parameters, you can check its `.formal` attribute.

    Args:
      annot: An annotation.
      seen: A seen set.
    """
    seen = seen or set()
    if annot in seen:
      return []
    if isinstance(annot, mixin.NestedAnnotation):
      # We track parameterized classes to avoid recursion errors when a class
      # contains itself.
      seen = seen | {annot}
    if isinstance(annot, abstract.TypeParameter):
      return [annot]
    elif isinstance(annot, abstract.TupleClass):
      annots = []
      for idx in range(annot.tuple_length):
        annots.extend(self.get_type_parameters(
            annot.formal_type_parameters[idx], seen))
      return annots
    elif isinstance(annot, mixin.NestedAnnotation):
      return sum((self.get_type_parameters(t, seen)
                  for _, t in annot.get_inner_types()), [])
    return []

  def get_callable_type_parameter_names(self, val: abstract.BaseValue):
    """Gets all TypeParameter names that appear in a Callable in 'val'."""
    type_params = set()
    seen = set()
    stack = [val]
    while stack:
      annot = stack.pop()
      if annot in seen:
        continue
      seen.add(annot)
      if annot.full_name == "typing.Callable":
        params = collections.Counter(self.get_type_parameters(annot))
        if isinstance(annot, abstract.CallableClass):
          # pytype represents Callable[[T1, T2], None] as
          # CallableClass({0: T1, 1: T2, ARGS: Union[T1, T2], RET: None}),
          # so we have to fix double-counting of argument type parameters.
          params -= collections.Counter(self.get_type_parameters(
              annot.formal_type_parameters[abstract_utils.ARGS]))
        # Type parameters that appear only once in a function signature are
        # invalid, so ignore them.
        type_params.update(p.name for p, n in params.items() if n > 1)
      elif isinstance(annot, mixin.NestedAnnotation):
        stack.extend(v for _, v in annot.get_inner_types())
    return type_params

  def convert_function_type_annotation(self, name, typ):
    visible = typ.data
    if len(visible) > 1:
      self.ctx.errorlog.ambiguous_annotation(self.ctx.vm.frames, visible, name)
      return None
    else:
      return visible[0]

  def convert_function_annotations(self, node, raw_annotations):
    """Convert raw annotations to a {name: annotation} dict."""
    if raw_annotations:
      # {"i": int, "return": str} is stored as (int, str, ("i", "return"))
      names = abstract_utils.get_atomic_python_constant(raw_annotations[-1])
      type_list = raw_annotations[:-1]
      annotations_list = []
      for name, t in zip(names, type_list):
        name = abstract_utils.get_atomic_python_constant(name)
        t = self.convert_function_type_annotation(name, t)
        annotations_list.append((name, t))
      return self.convert_annotations_list(node, annotations_list)
    else:
      return {}

  def convert_annotations_list(self, node, annotations_list):
    """Convert a (name, raw_annot) list to a {name: annotation} dict."""
    annotations = {}
    for name, t in annotations_list:
      if t is None or abstract_utils.is_ellipsis(t):
        # '...' is an experimental "inferred type"; see b/213607272.
        continue
      annot = self._process_one_annotation(node, t, name,
                                           self.ctx.vm.simple_stack())
      if annot is not None:
        annotations[name] = annot
    return annotations

  def convert_class_annotations(self, node, raw_annotations):
    """Convert a name -> raw_annot dict to annotations."""
    annotations = {}
    raw_items = raw_annotations.items()
    for name, t in raw_items:
      # Don't use the parameter name, since it's often something unhelpful
      # like `0`.
      annot = self._process_one_annotation(node, t, None,
                                           self.ctx.vm.simple_stack())
      annotations[name] = annot or self.ctx.convert.unsolvable
    return annotations

  def init_annotation(self, node, name, annot, container=None, extra_key=None):
    node, value = self.ctx.vm.init_class(
        node, annot, container=container, extra_key=extra_key)
    for d in value.data:
      d.from_annotation = name
    return node, value

  def extract_and_init_annotation(self, node, name, var):
    """Extracts an annotation from var and instantiates it."""
    frame = self.ctx.vm.frame
    substs = frame.substs
    if frame.func and isinstance(frame.func.data, abstract.BoundFunction):
      self_var = frame.first_arg
      if self_var:
        # self_var is an instance of (a subclass of) the class on which
        # frame.func is defined. We walk self_var's class's MRO to find the
        # defining class and grab its type parameter substitutions.
        defining_cls_name, _, _ = frame.func.data.name.rpartition(".")
        type_params = []
        defining_classes = []
        for v in self_var.data:
          for cls in v.cls.mro:
            if cls.name == defining_cls_name:
              # Normalize type parameter names by dropping the scope.
              type_params.extend(p.with_module(None) for p in cls.template)
              defining_classes.append(cls)
              break
        self_substs = tuple(
            abstract_utils.get_type_parameter_substitutions(cls, type_params)
            for cls in defining_classes)
        substs = abstract_utils.combine_substs(substs, self_substs)
    typ = self.extract_annotation(
        node,
        var,
        name,
        self.ctx.vm.simple_stack(),
        allowed_type_params=set(itertools.chain(*substs)))
    orig_typ = typ
    if isinstance(typ, abstract.FinalAnnotation):
      typ = typ.annotation
    elif isinstance(typ, typing_overlay.Final):
      typ = self.ctx.convert.unsolvable
    if typ.formal:
      typ = self.sub_one_annotation(node, typ, substs,
                                    instantiate_unbound=False)
    _, value = self.init_annotation(node, name, typ)
    return orig_typ, value

  def apply_annotation(self, node, op, name, value):
    """If there is an annotation for the op, return its value."""
    assert op is self.ctx.vm.frame.current_opcode
    if op.code.co_filename != self.ctx.vm.filename:
      return AnnotatedValue(None, value)
    if not op.annotation:
      return AnnotatedValue(None, value)
    annot = op.annotation
    if annot == "...":
      # Experimental "inferred type": see b/213607272.
      return AnnotatedValue(None, value)
    frame = self.ctx.vm.frame
    stack = self.ctx.vm.simple_stack()
    with self.ctx.vm.generate_late_annotations(stack):
      var, errorlog = abstract_utils.eval_expr(self.ctx, node, frame.f_globals,
                                               frame.f_locals, annot)
    if errorlog:
      self.ctx.errorlog.invalid_annotation(
          self.ctx.vm.frames, annot, details=errorlog.details)
    typ, annot_val = self.extract_and_init_annotation(node, name, var)
    if isinstance(typ, typing_overlay.Final):
      # return the original value, we want type inference here.
      return AnnotatedValue(None, value, final=True)
    elif isinstance(typ, abstract.FinalAnnotation):
      return AnnotatedValue(typ.annotation, annot_val, final=True)
    elif typ.full_name == "typing.TypeAlias":
      # Validate that 'value' is a legal type alias and use it.
      annot = self.extract_annotation(node, value, name, stack)
      return AnnotatedValue(None, annot.to_variable(node))
    else:
      return AnnotatedValue(typ, annot_val)

  def extract_annotation(
      self, node, var, name, stack,
      allowed_type_params: Optional[AbstractSet[str]] = None):
    """Returns an annotation extracted from 'var'.

    Args:
      node: The current node.
      var: The variable to extract from.
      name: The annotated name.
      stack: The frame stack.
      allowed_type_params: Type parameters that are allowed to appear in the
        annotation. 'None' means all are allowed. If non-None, the result of
        calling get_callable_type_parameter_names on the extracted annotation is
        also added to the allowed set.
    """
    try:
      typ = abstract_utils.get_atomic_value(var)
    except abstract_utils.ConversionError:
      self.ctx.errorlog.ambiguous_annotation(self.ctx.vm.frames, None, name)
      return self.ctx.convert.unsolvable
    typ = self._process_one_annotation(node, typ, name, stack)
    if not typ:
      return self.ctx.convert.unsolvable
    if typ.formal and allowed_type_params is not None:
      allowed_type_params = (allowed_type_params |
                             self.get_callable_type_parameter_names(typ))
      illegal_params = [x.name for x in self.get_type_parameters(typ)
                        if x.name not in allowed_type_params]
      if illegal_params:
        details = "TypeVar(s) %s not in scope" % ", ".join(
            repr(p) for p in utils.unique_list(illegal_params))
        if self.ctx.vm.frame.func:
          method = self.ctx.vm.frame.func.data
          if isinstance(method, abstract.BoundFunction):
            desc = "class"
            frame_name = method.name.rsplit(".", 1)[0]
          else:
            desc = "class" if method.is_class_builder else "method"
            frame_name = method.name
          details += f" for {desc} {frame_name!r}"
        if "AnyStr" in illegal_params:
          str_type = "Union[str, bytes]"
          details += (f"\nNote: For all string types, use {str_type}.")
        self.ctx.errorlog.invalid_annotation(stack, typ, details, name)
        return self.ctx.convert.unsolvable
    return typ

  def eval_multi_arg_annotation(self, node, func, annot, stack):
    """Evaluate annotation for multiple arguments (from a type comment)."""
    args, errorlog = self._eval_expr_as_tuple(node, annot, stack)
    if errorlog:
      self.ctx.errorlog.invalid_function_type_comment(
          stack, annot, details=errorlog.details)
    code = func.code
    expected = code.get_arg_count()
    names = code.co_varnames

    # This is a hack.  Specifying the type of the first arg is optional in
    # class and instance methods.  There is no way to tell at this time
    # how the function will be used, so if the first arg is self or cls we
    # make it optional.  The logic is somewhat convoluted because we don't
    # want to count the skipped argument in an error message.
    if len(args) != expected:
      if expected and names[0] in ["self", "cls"]:
        expected -= 1
        names = names[1:]

    if len(args) != expected:
      self.ctx.errorlog.invalid_function_type_comment(
          stack,
          annot,
          details="Expected %d args, %d given" % (expected, len(args)))
      return
    for name, arg in zip(names, args):
      resolved = self._process_one_annotation(node, arg, name, stack)
      if resolved is not None:
        func.signature.set_annotation(name, resolved)

  def _process_one_annotation(self, node, annotation, name, stack):
    """Change annotation / record errors where required."""
    # Make sure we pass in a frozen snapshot of the frame stack, rather than the
    # actual stack, since late annotations need to snapshot the stack at time of
    # creation in order to get the right line information for error messages.
    assert isinstance(stack, tuple), "stack must be an immutable sequence"

    if isinstance(annotation, abstract.AnnotationContainer):
      annotation = annotation.base_cls

    if isinstance(annotation, typing_overlay.Union):
      self.ctx.errorlog.invalid_annotation(stack, annotation, "Needs options",
                                           name)
      return None
    elif (name is not None and name != "return"
          and isinstance(annotation, typing_overlay.NoReturn)):
      self.ctx.errorlog.invalid_annotation(stack, annotation,
                                           "NoReturn is not allowed", name)
      return None
    elif (isinstance(annotation, abstract.Instance) and
          annotation.cls == self.ctx.convert.str_type):
      # String annotations : Late evaluation
      if isinstance(annotation, mixin.PythonConstant):
        expr = annotation.pyval
        if not expr:
          self.ctx.errorlog.invalid_annotation(stack, annotation,
                                               "Cannot be an empty string",
                                               name)
          return None
        frame = self.ctx.vm.frame
        # Immediately try to evaluate the reference, generating LateAnnotation
        # objects as needed. We don't store the entire string as a
        # LateAnnotation because:
        # - With __future__.annotations, all annotations look like forward
        #   references - most of them don't need to be late evaluated.
        # - Given an expression like "Union[str, NotYetDefined]", we want to
        #   evaluate the union immediately so we don't end up with a complex
        #   LateAnnotation, which can lead to bugs when instantiated.
        with self.ctx.vm.generate_late_annotations(stack):
          v, errorlog = abstract_utils.eval_expr(self.ctx, node,
                                                 frame.f_globals,
                                                 frame.f_locals, expr)
        if errorlog:
          self.ctx.errorlog.copy_from(errorlog.errors, stack)
        if len(v.data) == 1:
          return self._process_one_annotation(node, v.data[0], name, stack)
      self.ctx.errorlog.ambiguous_annotation(stack, [annotation], name)
      return None
    elif annotation.cls == self.ctx.convert.none_type:
      # PEP 484 allows to write "NoneType" as "None"
      return self.ctx.convert.none_type
    elif isinstance(annotation, mixin.NestedAnnotation):
      if annotation.processed:
        return annotation
      annotation.processed = True
      for key, typ in annotation.get_inner_types():
        processed = self._process_one_annotation(node, typ, name, stack)
        if processed is None:
          return None
        elif isinstance(processed, typing_overlay.NoReturn):
          self.ctx.errorlog.invalid_annotation(
              stack, typ, "NoReturn is not allowed as inner type", name)
          return None
        annotation.update_inner_type(key, processed)
      return annotation
    elif isinstance(annotation, (abstract.Class,
                                 abstract.AMBIGUOUS_OR_EMPTY,
                                 abstract.TypeParameter,
                                 abstract.FinalAnnotation,
                                 typing_overlay.Final,
                                 typing_overlay.NoReturn)):
      return annotation
    else:
      self.ctx.errorlog.invalid_annotation(stack, annotation, "Not a type",
                                           name)
      return None

  def _eval_expr_as_tuple(self, node, expr, stack):
    """Evaluate an expression as a tuple."""
    if not expr:
      return (), None

    f_globals = self.ctx.vm.frame.f_globals
    f_locals = self.ctx.vm.frame.f_locals
    with self.ctx.vm.generate_late_annotations(stack):
      result_var, errorlog = abstract_utils.eval_expr(self.ctx, node, f_globals,
                                                      f_locals, expr)
    result = abstract_utils.get_atomic_value(result_var)
    # If the result is a tuple, expand it.
    if (isinstance(result, mixin.PythonConstant) and
        isinstance(result.pyval, tuple)):
      return (tuple(abstract_utils.get_atomic_value(x) for x in result.pyval),
              errorlog)
    else:
      return (result,), errorlog

  def deformalize(self, value):
    # TODO(rechen): Instead of doing this, call sub_one_annotation() to replace
    # type parameters with their bound/constraints.
    while value.formal:
      if isinstance(value, abstract.ParameterizedClass):
        value = value.base_cls
      else:
        value = self.ctx.convert.unsolvable
    return value



from typing import Any

from openpyxl.descriptors.serialisable import Serialisable

class DataTable(Serialisable):
    tagname: str
    showHorzBorder: Any
    showVertBorder: Any
    showOutline: Any
    showKeys: Any
    spPr: Any
    graphicalProperties: Any
    txPr: Any
    extLst: Any
    __elements__: Any
    def __init__(
        self,
        showHorzBorder: Any | None = ...,
        showVertBorder: Any | None = ...,
        showOutline: Any | None = ...,
        showKeys: Any | None = ...,
        spPr: Any | None = ...,
        txPr: Any | None = ...,
        extLst: Any | None = ...,
    ) -> None: ...

class PlotArea(Serialisable):
    tagname: str
    layout: Any
    dTable: Any
    spPr: Any
    graphicalProperties: Any
    extLst: Any
    areaChart: Any
    area3DChart: Any
    lineChart: Any
    line3DChart: Any
    stockChart: Any
    radarChart: Any
    scatterChart: Any
    pieChart: Any
    pie3DChart: Any
    doughnutChart: Any
    barChart: Any
    bar3DChart: Any
    ofPieChart: Any
    surfaceChart: Any
    surface3DChart: Any
    bubbleChart: Any
    valAx: Any
    catAx: Any
    dateAx: Any
    serAx: Any
    __elements__: Any
    def __init__(
        self,
        layout: Any | None = ...,
        dTable: Any | None = ...,
        spPr: Any | None = ...,
        _charts=...,
        _axes=...,
        extLst: Any | None = ...,
    ) -> None: ...
    def to_tree(self, tagname: Any | None = ..., idx: Any | None = ..., namespace: Any | None = ...): ...
    @classmethod
    def from_tree(cls, node): ...



def list_or_args(keys, args): ...
def nativestr(x): ...
def delist(x): ...
def parse_to_list(response): ...
def parse_list_to_dict(response): ...
def parse_to_dict(response): ...
def random_string(length: int = ...) -> str: ...
def quote_string(v): ...
def decodeDictKeys(obj): ...
def stringify_param_value(value): ...



"""Function definitions in pyi files."""

import dataclasses

from typing import Any, Dict, Iterable, List, Optional, Tuple

from pytype.pytd import pytd


class OverloadedDecoratorError(Exception):
  """Inconsistent decorators on an overloaded function."""

  def __init__(self, name, typ):
    msg = "Overloaded signatures for {} disagree on {}decorators".format(
        name, (typ + " " if typ else ""))
    super().__init__(msg)


class PropertyDecoratorError(Exception):
  """Inconsistent property decorators on an overloaded function."""

  def __init__(self, name):
    msg = (f"Invalid property decorators for method `{name}` "
           "(need at most one each of @property, "
           f"@{name}.setter and @{name}.deleter)")
    super().__init__(msg)


@dataclasses.dataclass
class Param:
  """Internal representation of function parameters."""

  name: str
  type: Optional[pytd.Type] = None
  default: Any = None
  kind: pytd.ParameterKind = pytd.ParameterKind.REGULAR

  def to_pytd(self) -> pytd.Parameter:
    """Return a pytd.Parameter object for a normal argument."""
    if self.default is not None:
      default_type = self.default
      if self.type is None and default_type != pytd.NamedType("NoneType"):
        self.type = default_type
    if self.type is None:
      self.type = pytd.AnythingType()

    optional = self.default is not None
    return pytd.Parameter(self.name, self.type, self.kind, optional, None)


@dataclasses.dataclass(frozen=True)
class NameAndSig:
  """Internal representation of function signatures."""

  name: str
  signature: pytd.Signature
  decorator: Optional[str] = None
  is_abstract: bool = False
  is_coroutine: bool = False
  is_final: bool = False
  is_overload: bool = False

  @classmethod
  def make(
      cls,
      name: str,
      args: List[Tuple[str, pytd.Type]],
      return_type: pytd.Type
  ) -> "NameAndSig":
    """Make a new NameAndSig from an argument list."""
    params = tuple(Param(n, t).to_pytd() for (n, t) in args)
    sig = pytd.Signature(params=params, return_type=return_type,
                         starargs=None, starstarargs=None,
                         exceptions=(), template=())
    return cls(name, sig)


def pytd_return_type(
    name: str,
    return_type: Optional[pytd.Type],
    is_async: bool
) -> pytd.Type:
  """Convert function return type to pytd."""
  if name == "__init__":
    if (return_type is None or
        isinstance(return_type, pytd.AnythingType)):
      ret = pytd.NamedType("NoneType")
    else:
      ret = return_type
  elif is_async:
    base = pytd.NamedType("typing.Coroutine")
    params = (pytd.AnythingType(), pytd.AnythingType(), return_type)
    ret = pytd.GenericType(base, params)
  elif return_type is None:
    ret = pytd.AnythingType()
  else:
    ret = return_type
  return ret


def pytd_default_star_param() -> pytd.Parameter:
  return pytd.Parameter(
      "args", pytd.NamedType("tuple"), pytd.ParameterKind.REGULAR, True, None)


def pytd_default_starstar_param() -> pytd.Parameter:
  return pytd.Parameter(
      "kwargs", pytd.NamedType("dict"), pytd.ParameterKind.REGULAR, True, None)


def pytd_star_param(name: str, annotation: pytd.Type) -> pytd.Parameter:
  """Return a pytd.Parameter for a *args argument."""
  if annotation is None:
    param_type = pytd.NamedType("tuple")
  else:
    param_type = pytd.GenericType(
        pytd.NamedType("tuple"), (annotation,))
  return pytd.Parameter(
      name, param_type, pytd.ParameterKind.REGULAR, True, None)


def pytd_starstar_param(
    name: str, annotation: pytd.Type
) -> pytd.Parameter:
  """Return a pytd.Parameter for a **kwargs argument."""
  if annotation is None:
    param_type = pytd.NamedType("dict")
  else:
    param_type = pytd.GenericType(
        pytd.NamedType("dict"), (pytd.NamedType("str"), annotation))
  return pytd.Parameter(
      name, param_type, pytd.ParameterKind.REGULAR, True, None)


def _make_param(attr: pytd.Constant) -> pytd.Parameter:
  return Param(name=attr.name, type=attr.type, default=attr.value).to_pytd()


def generate_init(fields: Iterable[pytd.Constant]) -> pytd.Function:
  """Build an __init__ method from pytd class constants."""
  self_arg = Param("self").to_pytd()
  params = (self_arg,) + tuple(_make_param(c) for c in fields)
  # We call this at 'runtime' rather than from the parser, so we need to use the
  # resolved type of None, rather than NamedType("NoneType")
  ret = pytd.ClassType("builtins.NoneType")
  sig = pytd.Signature(params=params, return_type=ret,
                       starargs=None, starstarargs=None,
                       exceptions=(), template=())
  return pytd.Function("__init__", (sig,), kind=pytd.MethodKind.METHOD)


# -------------------------------------------
# Method signature merging


@dataclasses.dataclass
class _Property:
  type: str
  arity: int


def _property_decorators(name: str) -> Dict[str, _Property]:
  """Generates the property decorators for a method name."""
  return {
      "property": _Property("getter", 1),
      (name + ".setter"): _Property("setter", 2),
      (name + ".deleter"): _Property("deleter", 1)
  }


@dataclasses.dataclass
class _Properties:
  """Function property decorators."""

  getter: Optional[pytd.Signature] = None
  setter: Optional[pytd.Signature] = None
  deleter: Optional[pytd.Signature] = None

  def set(self, prop, sig, name):
    assert hasattr(self, prop), prop
    if getattr(self, prop):
      raise PropertyDecoratorError(name)
    setattr(self, prop, sig)


@dataclasses.dataclass
class _DecoratedFunction:
  """A mutable builder for pytd.Function values."""

  name: str
  sigs: List[pytd.Signature]
  is_abstract: bool = False
  is_coroutine: bool = False
  is_final: bool = False
  decorator: Optional[str] = None
  properties: Optional[_Properties] = dataclasses.field(init=False)
  prop_names: Dict[str, _Property] = dataclasses.field(init=False)

  @classmethod
  def make(cls, fn: NameAndSig):
    return cls(
        name=fn.name,
        sigs=[fn.signature],
        is_abstract=fn.is_abstract,
        is_coroutine=fn.is_coroutine,
        is_final=fn.is_final,
        decorator=fn.decorator)

  def __post_init__(self):
    self.prop_names = _property_decorators(self.name)
    if self.decorator in self.prop_names:
      self.properties = _Properties()
      self.add_property(self.decorator, self.sigs[0])
    else:
      self.properties = None

  def add_property(self, decorator, sig):
    prop = self.prop_names[decorator]
    if prop.arity == len(sig.params):
      self.properties.set(prop.type, sig, self.name)
    else:
      raise TypeError("Property decorator @%s needs %d param(s), got %d" %
                      (decorator, prop.arity, len(sig.params)))

  def add_overload(self, fn: NameAndSig):
    """Add an overloaded signature to a function."""
    # Check for decorator consistency. Note that we currently limit pyi files to
    # one decorator per function, other than @abstractmethod and @coroutine
    # which are special-cased.
    if (self.properties and fn.decorator in self.prop_names):
      # For properties, we can have at most one of setter, getter and deleter,
      # and no other overloads
      self.add_property(fn.decorator, fn.signature)
      # For properties, it's fine if, e.g., the getter is abstract but the
      # setter is not, so we skip the @abstractmethod and  @coroutine
      # consistency checks.
      return
    elif self.decorator == fn.decorator:
      # For other decorators, we can have multiple overloads but they need to
      # all have the same decorator
      self.sigs.append(fn.signature)
    else:
      raise OverloadedDecoratorError(self.name, None)
    # @abstractmethod and @coroutine can be combined with other decorators, but
    # they need to be consistent for all overloads
    if self.is_abstract != fn.is_abstract:
      raise OverloadedDecoratorError(self.name, "abstractmethod")
    if self.is_coroutine != fn.is_coroutine:
      raise OverloadedDecoratorError(self.name, "coroutine")


def merge_method_signatures(
    name_and_sigs: List[NameAndSig],
    check_unhandled_decorator: bool = False
) -> List[pytd.Function]:
  """Group the signatures by name, turning each group into a function."""
  functions = {}
  for fn in name_and_sigs:
    if fn.name not in functions:
      functions[fn.name] = _DecoratedFunction.make(fn)
    else:
      functions[fn.name].add_overload(fn)
  methods = []
  for name, fn in functions.items():
    if name == "__new__" or fn.decorator == "staticmethod":
      kind = pytd.MethodKind.STATICMETHOD
    elif name == "__init_subclass__" or fn.decorator == "classmethod":
      kind = pytd.MethodKind.CLASSMETHOD
    elif fn.properties:
      kind = pytd.MethodKind.PROPERTY
      # If we have only setters and/or deleters, replace them with a single
      # method foo(...) -> Any, so that we infer a constant `foo: Any` even if
      # the original method signatures are all `foo(...) -> None`. (If we have a
      # getter we use its return type, but in the absence of a getter we want to
      # fall back on Any since we cannot say anything about what the setter sets
      # the type of foo to.)
      if fn.properties.getter:
        fn.sigs = [fn.properties.getter]
      else:
        sig = fn.properties.setter or fn.properties.deleter
        fn.sigs = [sig.Replace(return_type=pytd.AnythingType())]
    elif fn.decorator and check_unhandled_decorator:
      raise ValueError(f"Unhandled decorator: {fn.decorator}")
    else:
      # Other decorators do not affect the kind
      kind = pytd.MethodKind.METHOD
    flags = pytd.MethodFlag.NONE
    if fn.is_abstract:
      flags |= pytd.MethodFlag.ABSTRACT
    if fn.is_coroutine:
      flags |= pytd.MethodFlag.COROUTINE
    if fn.is_final:
      flags |= pytd.MethodFlag.FINAL
    methods.append(pytd.Function(name, tuple(fn.sigs), kind, flags))
  return methods



from typing import Any

change_table: Any

def modify_operation(
    dn, changes, auto_encode, schema: Any | None = ..., validator: Any | None = ..., check_names: bool = ...
): ...
def modify_request_to_dict(request): ...
def modify_response_to_dict(response): ...



from collections.abc import MutableMapping
from typing import Any, Callable, Iterable, Iterator, Sequence, Text, TypeVar, overload
from typing_extensions import SupportsIndex

from google.protobuf.descriptor import Descriptor
from google.protobuf.internal.message_listener import MessageListener
from google.protobuf.internal.python_message import GeneratedProtocolMessageType
from google.protobuf.internal.type_checkers import _ValueChecker
from google.protobuf.message import Message

_T = TypeVar("_T")
_K = TypeVar("_K", bound=bool | int | Text)
_ScalarV = TypeVar("_ScalarV", bound=bool | int | float | Text | bytes)
_MessageV = TypeVar("_MessageV", bound=Message)
_M = TypeVar("_M")

class BaseContainer(Sequence[_T]):
    def __init__(self, message_listener: MessageListener) -> None: ...
    def __len__(self) -> int: ...
    def __ne__(self, other: object) -> bool: ...
    def __hash__(self) -> int: ...
    def sort(self, *, key: Callable[[_T], Any] | None = ..., reverse: bool = ...) -> None: ...
    @overload
    def __getitem__(self, key: SupportsIndex) -> _T: ...
    @overload
    def __getitem__(self, key: slice) -> list[_T]: ...

class RepeatedScalarFieldContainer(BaseContainer[_ScalarV]):
    def __init__(self, message_listener: MessageListener, type_checker: _ValueChecker[_ScalarV]) -> None: ...
    def append(self, value: _ScalarV) -> None: ...
    def insert(self, key: int, value: _ScalarV) -> None: ...
    def extend(self, elem_seq: Iterable[_ScalarV] | None) -> None: ...
    def MergeFrom(self: _M, other: _M) -> None: ...
    def remove(self, elem: _ScalarV) -> None: ...
    def pop(self, key: int = ...) -> _ScalarV: ...
    @overload
    def __setitem__(self, key: int, value: _ScalarV) -> None: ...
    @overload
    def __setitem__(self, key: slice, value: Iterable[_ScalarV]) -> None: ...
    def __getslice__(self, start: int, stop: int) -> list[_ScalarV]: ...
    def __setslice__(self, start: int, stop: int, values: Iterable[_ScalarV]) -> None: ...
    def __delitem__(self, key: int | slice) -> None: ...
    def __delslice__(self, start: int, stop: int) -> None: ...
    def __eq__(self, other: object) -> bool: ...

class RepeatedCompositeFieldContainer(BaseContainer[_MessageV]):
    def __init__(self, message_listener: MessageListener, message_descriptor: Descriptor) -> None: ...
    def add(self, **kwargs: Any) -> _MessageV: ...
    def append(self, value: _MessageV) -> None: ...
    def insert(self, key: int, value: _MessageV) -> None: ...
    def extend(self, elem_seq: Iterable[_MessageV]) -> None: ...
    def MergeFrom(self: _M, other: _M) -> None: ...
    def remove(self, elem: _MessageV) -> None: ...
    def pop(self, key: int = ...) -> _MessageV: ...
    def __getslice__(self, start: int, stop: int) -> list[_MessageV]: ...
    def __delitem__(self, key: int | slice) -> None: ...
    def __delslice__(self, start: int, stop: int) -> None: ...
    def __eq__(self, other: object) -> bool: ...

class ScalarMap(MutableMapping[_K, _ScalarV]):
    def __init__(
        self,
        message_listener: MessageListener,
        key_checker: _ValueChecker[_K],
        value_checker: _ValueChecker[_ScalarV],
        entry_descriptor: Descriptor,
    ) -> None: ...
    def __setitem__(self, k: _K, v: _ScalarV) -> None: ...
    def __delitem__(self, v: _K) -> None: ...
    def __getitem__(self, k: _K) -> _ScalarV: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[_K]: ...
    def __eq__(self, other: object) -> bool: ...
    @overload
    def get(self, key: _K, default: None = ...) -> _ScalarV: ...
    @overload
    def get(self, key: _K, default: _ScalarV | _T) -> _ScalarV | _T: ...
    def MergeFrom(self: _M, other: _M): ...
    def InvalidateIterators(self) -> None: ...
    def GetEntryClass(self) -> GeneratedProtocolMessageType: ...

class MessageMap(MutableMapping[_K, _MessageV]):
    def __init__(
        self,
        message_listener: MessageListener,
        message_descriptor: Descriptor,
        key_checker: _ValueChecker[_K],
        entry_descriptor: Descriptor,
    ) -> None: ...
    def __setitem__(self, k: _K, v: _MessageV) -> None: ...
    def __delitem__(self, v: _K) -> None: ...
    def __getitem__(self, k: _K) -> _MessageV: ...
    def __len__(self) -> int: ...
    def __iter__(self) -> Iterator[_K]: ...
    def __eq__(self, other: object) -> bool: ...
    @overload
    def get(self, key: _K, default: None = ...) -> _MessageV: ...
    @overload
    def get(self, key: _K, default: _MessageV | _T) -> _MessageV | _T: ...
    def get_or_create(self, key: _K) -> _MessageV: ...
    def MergeFrom(self: _M, other: _M): ...
    def InvalidateIterators(self) -> None: ...
    def GetEntryClass(self) -> GeneratedProtocolMessageType: ...



from _typeshed import StrPath
from collections.abc import Iterator
from typing import Match

class Store:
    def __init__(
        self,
        gpg_bin: str = ...,
        git_bin: str = ...,
        store_dir: str = ...,
        use_agent: bool = ...,
        interactive: bool = ...,
        verbose: bool = ...,
    ) -> None: ...
    def __iter__(self) -> Iterator[str]: ...
    def is_init(self) -> bool: ...
    def init_store(self, gpg_ids: None | str | list[str], path: StrPath | None = ...) -> None: ...
    def init_git(self) -> None: ...
    def git(self, method: str, *args: object, **kwargs: object) -> None: ...
    def get_key(self, path: StrPath | None) -> str | None: ...
    def set_key(self, path: StrPath | None, key_data: str, force: bool = ...) -> None: ...
    def remove_path(self, path: StrPath, recursive: bool = ..., force: bool = ...) -> None: ...
    def gen_key(
        self, path: StrPath | None, length: int, symbols: bool = ..., force: bool = ..., inplace: bool = ...
    ) -> str | None: ...
    def copy_path(self, old_path: StrPath, new_path: StrPath, force: bool = ...) -> None: ...
    def move_path(self, old_path: StrPath, new_path: StrPath, force: bool = ...) -> None: ...
    def list_dir(self, path: StrPath) -> tuple[list[str], list[str]]: ...
    def iter_dir(self, path: StrPath) -> Iterator[str]: ...
    def find(self, names: None | str | list[str]) -> list[str]: ...
    def search(self, term: str) -> dict[str, list[tuple[str, Match[str]]]]: ...



from typing import Any

def guess_content_type(filename, default=...): ...
def format_header_param(name, value): ...

class RequestField:
    data: Any
    headers: Any
    def __init__(self, name, data, filename=..., headers=...) -> None: ...
    @classmethod
    def from_tuples(cls, fieldname, value): ...
    def render_headers(self): ...
    def make_multipart(self, content_disposition=..., content_type=..., content_location=...): ...



from email.message import Message
from typing import IO
from urllib.response import addinfourl

__all__ = ["URLError", "HTTPError", "ContentTooShortError"]

class URLError(IOError):
    reason: str | BaseException
    def __init__(self, reason: str | BaseException, filename: str | None = ...) -> None: ...

class HTTPError(URLError, addinfourl):
    @property
    def reason(self) -> str: ...  # type: ignore[override]
    code: int
    def __init__(self, url: str, code: int, msg: str, hdrs: Message, fp: IO[bytes] | None) -> None: ...

class ContentTooShortError(URLError):
    content: tuple[str, Message]
    def __init__(self, message: str, content: tuple[str, Message]) -> None: ...



from collections.abc import Iterable, Iterator
from typing import Any, TypeVar, overload

CallExpr = Any  # from mypy.nodes
Context = Any  # from mypy.nodes
Expression = Any  # from mypy.nodes
JsonDict = Any  # from mypy.nodes
NameExpr = Any  # from mypy.nodes
Statement = Any  # from mypy.nodes
TypeInfo = Any  # from mypy.nodes
AttributeContext = Any  # from mypy.plugin
ClassDefContext = Any  # from mypy.plugin
DynamicClassDefContext = Any  # from mypy.plugin
SemanticAnalyzerPluginInterface = Any  # from mypy.plugin
Type = Any  # from mypy.types

_TArgType = TypeVar("_TArgType", bound=CallExpr | NameExpr)

class SQLAlchemyAttribute:
    name: Any
    line: Any
    column: Any
    type: Any
    info: Any
    def __init__(self, name: str, line: int, column: int, typ: Type | None, info: TypeInfo) -> None: ...
    def serialize(self) -> JsonDict: ...
    def expand_typevar_from_subtype(self, sub_type: TypeInfo) -> None: ...
    @classmethod
    def deserialize(cls, info: TypeInfo, data: JsonDict, api: SemanticAnalyzerPluginInterface) -> SQLAlchemyAttribute: ...

def name_is_dunder(name): ...
def establish_as_sqlalchemy(info: TypeInfo) -> None: ...
def set_is_base(info: TypeInfo) -> None: ...
def get_is_base(info: TypeInfo) -> bool: ...
def has_declarative_base(info: TypeInfo) -> bool: ...
def set_has_table(info: TypeInfo) -> None: ...
def get_has_table(info: TypeInfo) -> bool: ...
def get_mapped_attributes(info: TypeInfo, api: SemanticAnalyzerPluginInterface) -> list[SQLAlchemyAttribute] | None: ...
def set_mapped_attributes(info: TypeInfo, attributes: list[SQLAlchemyAttribute]) -> None: ...
def fail(api: SemanticAnalyzerPluginInterface, msg: str, ctx: Context) -> None: ...
def add_global(ctx: ClassDefContext | DynamicClassDefContext, module: str, symbol_name: str, asname: str) -> None: ...
@overload
def get_callexpr_kwarg(callexpr: CallExpr, name: str, *, expr_types: None = ...) -> CallExpr | NameExpr | None: ...
@overload
def get_callexpr_kwarg(callexpr: CallExpr, name: str, *, expr_types: tuple[type[_TArgType], ...]) -> _TArgType | None: ...
def flatten_typechecking(stmts: Iterable[Statement]) -> Iterator[Statement]: ...
def unbound_to_instance(api: SemanticAnalyzerPluginInterface, typ: Type) -> Type: ...
def info_for_cls(cls, api: SemanticAnalyzerPluginInterface) -> TypeInfo | None: ...
def expr_to_mapped_constructor(expr: Expression) -> CallExpr: ...



"""Compiles a single .py to a .pyc and writes it to stdout."""

# These are C modules built into Python. Don't add any modules that are
# implemented in a .py:
import importlib.util
import marshal
import re
import sys

MAGIC = importlib.util.MAGIC_NUMBER

# This pattern is as per PEP-263.
ENCODING_PATTERN = "^[ \t\v]*#.*?coding[:=][ \t]*([-_.a-zA-Z0-9]+)"


def is_comment_only(line):
  return re.match("[ \t\v]*#.*", line) is not None


def _write32(f, w):
  f.write(bytearray([
      (w >> 0) & 0xff,
      (w >> 8) & 0xff,
      (w >> 16) & 0xff,
      (w >> 24) & 0xff]))


def write_pyc(f, codeobject, source_size=0, timestamp=0):
  f.write(MAGIC)
  _write32(f, timestamp)
  _write32(f, source_size)
  f.write(marshal.dumps(codeobject))


def compile_to_pyc(data_file, filename, output, mode):
  """Compile the source code to byte code."""
  with open(data_file, encoding="utf-8") as fi:
    src = fi.read()
  compile_src_to_pyc(src, filename, output, mode)


def strip_encoding(src):
  """Strip encoding from a src string assumed to be read from a file."""
  # Python 2's compile function does not like the line specifying the encoding.
  # So, we strip it off if it is present, replacing it with an empty comment to
  # preserve the original line numbering. As per PEP-263, the line specifying
  # the encoding can occur only in the first or the second line.
  if "\n" not in src:
    return src
  l1, rest = src.split("\n", 1)
  if re.match(ENCODING_PATTERN, l1.rstrip()):
    return "#\n" + rest
  elif "\n" not in rest:
    return src
  l2, rest = rest.split("\n", 1)
  if is_comment_only(l1) and re.match(ENCODING_PATTERN, l2.rstrip()):
    return "#\n#\n" + rest
  return src


def compile_src_to_pyc(src, filename, output, mode):
  """Compile a string of source code."""
  try:
    codeobject = compile(src, filename, mode)
  except Exception as err:  # pylint: disable=broad-except
    output.write(b"\1")
    output.write(str(err).encode("utf-8"))
  else:
    output.write(b"\0")
    write_pyc(output, codeobject)


def main():
  if len(sys.argv) != 4:
    sys.exit(1)
  output = sys.stdout.buffer if hasattr(sys.stdout, "buffer") else sys.stdout
  compile_to_pyc(data_file=sys.argv[1], filename=sys.argv[2],
                 output=output, mode=sys.argv[3])


if __name__ == "__main__":
  main()



from typing import Any

class DesiredCapabilities:
    FIREFOX: Any
    INTERNETEXPLORER: Any
    EDGE: Any
    CHROME: Any
    OPERA: Any
    SAFARI: Any
    HTMLUNIT: Any
    HTMLUNITWITHJS: Any
    IPHONE: Any
    IPAD: Any
    ANDROID: Any
    PHANTOMJS: Any
    WEBKITGTK: Any



from typing import Any

proxies: Any

def manage(module, **params): ...
def clear_managers() -> None: ...

class _DBProxy:
    module: Any
    kw: Any
    poolclass: Any
    pools: Any
    def __init__(self, module, poolclass=..., **kw) -> None: ...
    def close(self) -> None: ...
    def __del__(self) -> None: ...
    def __getattr__(self, key): ...
    def get_pool(self, *args, **kw): ...
    def connect(self, *args, **kw): ...
    def dispose(self, *args, **kw) -> None: ...



from typing import Any

from openpyxl.descriptors.serialisable import Serialisable

class SmartTag(Serialisable):
    tagname: str
    namespaceUri: Any
    name: Any
    url: Any
    def __init__(self, namespaceUri: Any | None = ..., name: Any | None = ..., url: Any | None = ...) -> None: ...

class SmartTagList(Serialisable):
    tagname: str
    smartTagType: Any
    __elements__: Any
    def __init__(self, smartTagType=...) -> None: ...

class SmartTagProperties(Serialisable):
    tagname: str
    embed: Any
    show: Any
    def __init__(self, embed: Any | None = ..., show: Any | None = ...) -> None: ...



from typing import IO

_encodable = bytes | unicode
_decodable = bytes | unicode

def b64encode(s: _encodable, altchars: bytes | None = ...) -> bytes: ...
def b64decode(s: _decodable, altchars: bytes | None = ..., validate: bool = ...) -> bytes: ...
def standard_b64encode(s: _encodable) -> bytes: ...
def standard_b64decode(s: _decodable) -> bytes: ...
def urlsafe_b64encode(s: _encodable) -> bytes: ...
def urlsafe_b64decode(s: _decodable) -> bytes: ...
def b32encode(s: _encodable) -> bytes: ...
def b32decode(s: _decodable, casefold: bool = ..., map01: bytes | None = ...) -> bytes: ...
def b16encode(s: _encodable) -> bytes: ...
def b16decode(s: _decodable, casefold: bool = ...) -> bytes: ...
def decode(input: IO[bytes], output: IO[bytes]) -> None: ...
def encode(input: IO[bytes], output: IO[bytes]) -> None: ...
def encodestring(s: bytes) -> bytes: ...
def decodestring(s: bytes) -> bytes: ...



from logging import Logger
from socket import socket
from threading import Condition, Event, Lock, Thread
from types import ModuleType
from typing import Any, Callable, Iterable, Protocol, Sequence

from paramiko.auth_handler import AuthHandler, _InteractiveCallback
from paramiko.channel import Channel
from paramiko.message import Message
from paramiko.packet import Packetizer
from paramiko.pkey import PKey
from paramiko.server import ServerInterface, SubsystemHandler
from paramiko.sftp_client import SFTPClient
from paramiko.ssh_gss import _SSH_GSSAuth
from paramiko.util import ClosingContextManager

_Addr = tuple[str, int]
_SocketLike = str | _Addr | socket | Channel

class _KexEngine(Protocol):
    def start_kex(self) -> None: ...
    def parse_next(self, ptype: int, m: Message) -> None: ...

class Transport(Thread, ClosingContextManager):
    active: bool
    hostname: str | None
    sock: socket | Channel
    packetizer: Packetizer
    local_version: str
    remote_version: str
    local_cipher: str
    local_kex_init: bytes | None
    local_mac: str | None
    local_compression: str | None
    session_id: bytes | None
    host_key_type: str | None
    host_key: PKey | None
    use_gss_kex: bool
    gss_kex_used: bool
    kexgss_ctxt: _SSH_GSSAuth | None
    gss_host: str
    kex_engine: _KexEngine | None
    H: bytes | None
    K: int | None
    initial_kex_done: bool
    in_kex: bool
    authenticated: bool
    lock: Lock
    channel_events: dict[int, Event]
    channels_seen: dict[int, bool]
    default_max_packet_size: int
    default_window_size: int
    saved_exception: Exception | None
    clear_to_send: Event
    clear_to_send_lock: Lock
    clear_to_send_timeout: float
    log_name: str
    logger: Logger
    auth_handler: AuthHandler | None
    global_response: Message | None
    completion_event: Event | None
    banner_timeout: float
    handshake_timeout: float
    auth_timeout: float
    disabled_algorithms: dict[str, Iterable[str]] | None
    server_mode: bool
    server_object: ServerInterface | None
    server_key_dict: dict[str, PKey]
    server_accepts: list[Channel]
    server_accept_cv: Condition
    subsystem_table: dict[str, tuple[type[SubsystemHandler], tuple[Any, ...], dict[str, Any]]]
    sys: ModuleType
    def __init__(
        self,
        sock: _SocketLike,
        default_window_size: int = ...,
        default_max_packet_size: int = ...,
        gss_kex: bool = ...,
        gss_deleg_creds: bool = ...,
        disabled_algorithms: dict[str, Iterable[str]] | None = ...,
    ) -> None: ...
    @property
    def preferred_ciphers(self) -> Sequence[str]: ...
    @property
    def preferred_macs(self) -> Sequence[str]: ...
    @property
    def preferred_keys(self) -> Sequence[str]: ...
    @property
    def preferred_kex(self) -> Sequence[str]: ...
    @property
    def preferred_compression(self) -> Sequence[str]: ...
    def atfork(self) -> None: ...
    def get_security_options(self) -> SecurityOptions: ...
    def set_gss_host(self, gss_host: str | None, trust_dns: bool = ..., gssapi_requested: bool = ...) -> None: ...
    def start_client(self, event: Event | None = ..., timeout: float | None = ...) -> None: ...
    def start_server(self, event: Event = ..., server: ServerInterface | None = ...) -> None: ...
    def add_server_key(self, key: PKey) -> None: ...
    def get_server_key(self) -> PKey | None: ...
    @staticmethod
    def load_server_moduli(filename: str | None = ...) -> bool: ...
    def close(self) -> None: ...
    def get_remote_server_key(self) -> PKey: ...
    def is_active(self) -> bool: ...
    def open_session(
        self, window_size: int | None = ..., max_packet_size: int | None = ..., timeout: float | None = ...
    ) -> Channel: ...
    def open_x11_channel(self, src_addr: _Addr = ...) -> Channel: ...
    def open_forward_agent_channel(self) -> Channel: ...
    def open_forwarded_tcpip_channel(self, src_addr: _Addr, dest_addr: _Addr) -> Channel: ...
    def open_channel(
        self,
        kind: str,
        dest_addr: _Addr | None = ...,
        src_addr: _Addr | None = ...,
        window_size: int | None = ...,
        max_packet_size: int | None = ...,
        timeout: float | None = ...,
    ) -> Channel: ...
    def request_port_forward(
        self, address: str, port: int, handler: Callable[[Channel, _Addr, _Addr], None] | None = ...
    ) -> int: ...
    def cancel_port_forward(self, address: str, port: int) -> None: ...
    def open_sftp_client(self) -> SFTPClient | None: ...
    def send_ignore(self, byte_count: int = ...) -> None: ...
    def renegotiate_keys(self) -> None: ...
    def set_keepalive(self, interval: int) -> None: ...
    def global_request(self, kind: str, data: Iterable[Any] | None = ..., wait: bool = ...) -> Message | None: ...
    def accept(self, timeout: float | None = ...) -> Channel | None: ...
    def connect(
        self,
        hostkey: PKey | None = ...,
        username: str = ...,
        password: str | None = ...,
        pkey: PKey | None = ...,
        gss_host: str | None = ...,
        gss_auth: bool = ...,
        gss_kex: bool = ...,
        gss_deleg_creds: bool = ...,
        gss_trust_dns: bool = ...,
    ) -> None: ...
    def get_exception(self) -> Exception | None: ...
    def set_subsystem_handler(self, name: str, handler: type[SubsystemHandler], *larg: Any, **kwarg: Any) -> None: ...
    def is_authenticated(self) -> bool: ...
    def get_username(self) -> str | None: ...
    def get_banner(self) -> bytes | None: ...
    def auth_none(self, username: str) -> list[str]: ...
    def auth_password(self, username: str, password: str, event: Event | None = ..., fallback: bool = ...) -> list[str]: ...
    def auth_publickey(self, username: str, key: PKey, event: Event | None = ...) -> list[str]: ...
    def auth_interactive(self, username: str, handler: _InteractiveCallback, submethods: str = ...) -> list[str]: ...
    def auth_interactive_dumb(
        self, username: str, handler: _InteractiveCallback | None = ..., submethods: str = ...
    ) -> list[str]: ...
    def auth_gssapi_with_mic(self, username: str, gss_host: str, gss_deleg_creds: bool) -> list[str]: ...
    def auth_gssapi_keyex(self, username: str) -> list[str]: ...
    def set_log_channel(self, name: str) -> None: ...
    def get_log_channel(self) -> str: ...
    def set_hexdump(self, hexdump: bool) -> None: ...
    def get_hexdump(self) -> bool: ...
    def use_compression(self, compress: bool = ...) -> None: ...
    def getpeername(self) -> tuple[str, int]: ...
    def stop_thread(self) -> None: ...
    def run(self) -> None: ...

class SecurityOptions:
    def __init__(self, transport: Transport) -> None: ...
    @property
    def ciphers(self) -> Sequence[str]: ...
    @ciphers.setter
    def ciphers(self, x: Sequence[str]) -> None: ...
    @property
    def digests(self) -> Sequence[str]: ...
    @digests.setter
    def digests(self, x: Sequence[str]) -> None: ...
    @property
    def key_types(self) -> Sequence[str]: ...
    @key_types.setter
    def key_types(self, x: Sequence[str]) -> None: ...
    @property
    def kex(self) -> Sequence[str]: ...
    @kex.setter
    def kex(self, x: Sequence[str]) -> None: ...
    @property
    def compression(self) -> Sequence[str]: ...
    @compression.setter
    def compression(self, x: Sequence[str]) -> None: ...

class ChannelMap:
    def __init__(self) -> None: ...
    def put(self, chanid: int, chan: Channel) -> None: ...
    def get(self, chanid: int) -> Channel: ...
    def delete(self, chanid: int) -> None: ...
    def values(self) -> list[Channel]: ...
    def __len__(self) -> int: ...



from typing import Any

log: Any
SIGNATURE_HMAC_SHA1: str
SIGNATURE_HMAC_SHA256: str
SIGNATURE_HMAC_SHA512: str
SIGNATURE_HMAC: str
SIGNATURE_RSA_SHA1: str
SIGNATURE_RSA_SHA256: str
SIGNATURE_RSA_SHA512: str
SIGNATURE_RSA: str
SIGNATURE_PLAINTEXT: str
SIGNATURE_METHODS: Any
SIGNATURE_TYPE_AUTH_HEADER: str
SIGNATURE_TYPE_QUERY: str
SIGNATURE_TYPE_BODY: str
CONTENT_TYPE_FORM_URLENCODED: str

class Client:
    SIGNATURE_METHODS: Any
    @classmethod
    def register_signature_method(cls, method_name, method_callback) -> None: ...
    client_key: Any
    client_secret: Any
    resource_owner_key: Any
    resource_owner_secret: Any
    signature_method: Any
    signature_type: Any
    callback_uri: Any
    rsa_key: Any
    verifier: Any
    realm: Any
    encoding: Any
    decoding: Any
    nonce: Any
    timestamp: Any
    def __init__(
        self,
        client_key,
        client_secret: Any | None = ...,
        resource_owner_key: Any | None = ...,
        resource_owner_secret: Any | None = ...,
        callback_uri: Any | None = ...,
        signature_method=...,
        signature_type=...,
        rsa_key: Any | None = ...,
        verifier: Any | None = ...,
        realm: Any | None = ...,
        encoding: str = ...,
        decoding: Any | None = ...,
        nonce: Any | None = ...,
        timestamp: Any | None = ...,
    ): ...
    def get_oauth_signature(self, request): ...
    def get_oauth_params(self, request): ...
    def sign(self, uri, http_method: str = ..., body: Any | None = ..., headers: Any | None = ..., realm: Any | None = ...): ...



from typing import NoReturn

def isPath(f: object) -> bool: ...
def isDirectory(f: object) -> bool: ...

class deferred_error:
    ex: BaseException
    def __init__(self, ex: BaseException) -> None: ...
    def __getattr__(self, elt: str) -> NoReturn: ...



class TouchActions:
    def __init__(self, driver) -> None: ...
    def perform(self) -> None: ...
    def tap(self, on_element): ...
    def double_tap(self, on_element): ...
    def tap_and_hold(self, xcoord, ycoord): ...
    def move(self, xcoord, ycoord): ...
    def release(self, xcoord, ycoord): ...
    def scroll(self, xoffset, yoffset): ...
    def scroll_from_element(self, on_element, xoffset, yoffset): ...
    def long_press(self, on_element): ...
    def flick(self, xspeed, yspeed): ...
    def flick_element(self, on_element, xoffset, yoffset, speed): ...
    def __enter__(self): ...
    def __exit__(self, _type, _value, _traceback) -> None: ...



from collections import OrderedDict
from typing import Any, Iterator

from .locale import Locale

LOCALE_SPLIT_PATTERN: Any

class LocaleDataLoader:
    def get_locale_map(
        self,
        languages: list[str] | None = ...,
        locales: list[str] | None = ...,
        region: str | None = ...,
        use_given_order: bool = ...,
        allow_conflicting_locales: bool = ...,
    ) -> OrderedDict[str, list[Any] | str | int]: ...
    def get_locales(
        self,
        languages: list[str] | None = ...,
        locales: list[str] | None = ...,
        region: str | None = ...,
        use_given_order: bool = ...,
        allow_conflicting_locales: bool = ...,
    ) -> Iterator[Locale]: ...
    def get_locale(self, shortname: str) -> Locale: ...

default_loader: Any



"""Utilities to deal with running subprocesses."""

import subprocess


class BinaryRun:
  """Convenience wrapper around subprocess.

  Use as:
    ret, out, err = BinaryRun([exe, arg, ...]).communicate()
  """

  def __init__(self, args, dry_run=False):
    self.args = args
    self.results = None

    if dry_run:
      self.results = (0, "", "")
    else:
      self.proc = subprocess.Popen(  # pylint: disable=consider-using-with
          self.args,
          stdout=subprocess.PIPE,
          stderr=subprocess.PIPE)

  def communicate(self):
    if self.results:
      # We are running in dry-run mode.
      return self.results

    stdout, stderr = self.proc.communicate()
    self.results = self.proc.returncode, stdout, stderr
    return self.results


def can_run(exe, *args):
  """Check if running exe with args works."""
  try:
    BinaryRun([exe] + list(args)).communicate()
    return True
  except OSError:
    return False



import sys
from _typeshed import Self
from typing import Any, BinaryIO, Generic, Iterable, MutableSequence, TypeVar, overload
from typing_extensions import Literal, SupportsIndex

_IntTypeCode = Literal["b", "B", "h", "H", "i", "I", "l", "L", "q", "Q"]
_FloatTypeCode = Literal["f", "d"]
_UnicodeTypeCode = Literal["u"]
_TypeCode = _IntTypeCode | _FloatTypeCode | _UnicodeTypeCode

_T = TypeVar("_T", int, float, str)

typecodes: str

class array(MutableSequence[_T], Generic[_T]):
    @property
    def typecode(self) -> _TypeCode: ...
    @property
    def itemsize(self) -> int: ...
    @overload
    def __init__(self: array[int], __typecode: _IntTypeCode, __initializer: bytes | Iterable[_T] = ...) -> None: ...
    @overload
    def __init__(self: array[float], __typecode: _FloatTypeCode, __initializer: bytes | Iterable[_T] = ...) -> None: ...
    @overload
    def __init__(self: array[str], __typecode: _UnicodeTypeCode, __initializer: bytes | Iterable[_T] = ...) -> None: ...
    @overload
    def __init__(self, typecode: str, __initializer: bytes | Iterable[_T] = ...) -> None: ...
    def append(self, __v: _T) -> None: ...
    def buffer_info(self) -> tuple[int, int]: ...
    def byteswap(self) -> None: ...
    def count(self, __v: _T) -> int: ...
    def extend(self, __bb: Iterable[_T]) -> None: ...
    def frombytes(self, __buffer: bytes) -> None: ...
    def fromfile(self, __f: BinaryIO, __n: int) -> None: ...
    def fromlist(self, __list: list[_T]) -> None: ...
    def fromunicode(self, __ustr: str) -> None: ...
    if sys.version_info >= (3, 10):
        def index(self, __v: _T, __start: int = ..., __stop: int = ...) -> int: ...
    else:
        def index(self, __v: _T) -> int: ...  # type: ignore[override]

    def insert(self, __i: int, __v: _T) -> None: ...
    def pop(self, __i: int = ...) -> _T: ...
    def remove(self, __v: _T) -> None: ...
    def reverse(self) -> None: ...
    def tobytes(self) -> bytes: ...
    def tofile(self, __f: BinaryIO) -> None: ...
    def tolist(self) -> list[_T]: ...
    def tounicode(self) -> str: ...
    if sys.version_info < (3, 9):
        def fromstring(self, __buffer: bytes) -> None: ...
        def tostring(self) -> bytes: ...

    def __contains__(self, __key: object) -> bool: ...
    def __len__(self) -> int: ...
    @overload
    def __getitem__(self, __i: SupportsIndex) -> _T: ...
    @overload
    def __getitem__(self, __s: slice) -> array[_T]: ...
    @overload  # type: ignore[override]
    def __setitem__(self, __i: SupportsIndex, __o: _T) -> None: ...
    @overload
    def __setitem__(self, __s: slice, __o: array[_T]) -> None: ...
    def __delitem__(self, __i: SupportsIndex | slice) -> None: ...
    def __add__(self, __x: array[_T]) -> array[_T]: ...
    def __ge__(self, __other: array[_T]) -> bool: ...
    def __gt__(self, __other: array[_T]) -> bool: ...
    def __iadd__(self: Self, __x: array[_T]) -> Self: ...  # type: ignore[override]
    def __imul__(self: Self, __n: int) -> Self: ...
    def __le__(self, __other: array[_T]) -> bool: ...
    def __lt__(self, __other: array[_T]) -> bool: ...
    def __mul__(self, __n: int) -> array[_T]: ...
    def __rmul__(self, __n: int) -> array[_T]: ...
    def __copy__(self) -> array[_T]: ...
    def __deepcopy__(self, __unused: Any) -> array[_T]: ...

ArrayType = array



from typing import Any

from ..orm import interfaces
from ..sql.operators import ColumnOperators
from ..util import memoized_property

def association_proxy(target_collection, attr, **kw): ...

ASSOCIATION_PROXY: Any

class AssociationProxy(interfaces.InspectionAttrInfo):
    is_attribute: bool
    extension_type: Any
    target_collection: Any
    value_attr: Any
    creator: Any
    getset_factory: Any
    proxy_factory: Any
    proxy_bulk_set: Any
    cascade_scalar_deletes: Any
    key: Any
    info: Any
    def __init__(
        self,
        target_collection,
        attr,
        creator: Any | None = ...,
        getset_factory: Any | None = ...,
        proxy_factory: Any | None = ...,
        proxy_bulk_set: Any | None = ...,
        info: Any | None = ...,
        cascade_scalar_deletes: bool = ...,
    ) -> None: ...
    def __get__(self, obj, class_): ...
    def __set__(self, obj, values): ...
    def __delete__(self, obj): ...
    def for_class(self, class_, obj: Any | None = ...): ...

class AssociationProxyInstance:
    parent: Any
    key: Any
    owning_class: Any
    target_collection: Any
    collection_class: Any
    target_class: Any
    value_attr: Any
    def __init__(self, parent, owning_class, target_class, value_attr) -> None: ...
    @classmethod
    def for_proxy(cls, parent, owning_class, parent_instance): ...
    def __clause_element__(self) -> None: ...
    @property
    def remote_attr(self): ...
    @property
    def local_attr(self): ...
    @property
    def attr(self): ...
    @memoized_property
    def scalar(self): ...
    @property
    def info(self): ...
    def get(self, obj): ...
    def set(self, obj, values) -> None: ...
    def delete(self, obj) -> None: ...
    def any(self, criterion: Any | None = ..., **kwargs): ...
    def has(self, criterion: Any | None = ..., **kwargs): ...

class AmbiguousAssociationProxyInstance(AssociationProxyInstance):
    def get(self, obj): ...
    def __eq__(self, obj): ...
    def __ne__(self, obj): ...
    def any(self, criterion: Any | None = ..., **kwargs) -> None: ...
    def has(self, criterion: Any | None = ..., **kwargs) -> None: ...

class ObjectAssociationProxyInstance(AssociationProxyInstance):
    def contains(self, obj): ...
    def __eq__(self, obj): ...
    def __ne__(self, obj): ...

class ColumnAssociationProxyInstance(ColumnOperators[Any], AssociationProxyInstance):
    def __eq__(self, other) -> ColumnOperators[Any]: ...  # type: ignore[override]
    def operate(self, op, *other, **kwargs): ...

class _lazy_collection:
    parent: Any
    target: Any
    def __init__(self, obj, target) -> None: ...
    def __call__(self): ...

class _AssociationCollection:
    lazy_collection: Any
    creator: Any
    getter: Any
    setter: Any
    parent: Any
    def __init__(self, lazy_collection, creator, getter, setter, parent) -> None: ...
    @property
    def col(self): ...
    def __len__(self): ...
    def __bool__(self): ...
    __nonzero__: Any

class _AssociationList(_AssociationCollection):
    def __getitem__(self, index): ...
    def __setitem__(self, index, value) -> None: ...
    def __delitem__(self, index) -> None: ...
    def __contains__(self, value): ...
    def __getslice__(self, start, end): ...
    def __setslice__(self, start, end, values) -> None: ...
    def __delslice__(self, start, end) -> None: ...
    def __iter__(self): ...
    def append(self, value) -> None: ...
    def count(self, value): ...
    def extend(self, values) -> None: ...
    def insert(self, index, value) -> None: ...
    def pop(self, index: int = ...): ...
    def remove(self, value) -> None: ...
    def reverse(self) -> None: ...
    def sort(self) -> None: ...
    def clear(self) -> None: ...
    def __eq__(self, other): ...
    def __ne__(self, other): ...
    def __lt__(self, other): ...
    def __le__(self, other): ...
    def __gt__(self, other): ...
    def __ge__(self, other): ...
    def __cmp__(self, other): ...
    def __add__(self, iterable): ...
    def __radd__(self, iterable): ...
    def __mul__(self, n): ...
    __rmul__: Any
    def __iadd__(self, iterable): ...
    def __imul__(self, n): ...
    def index(self, item, *args): ...
    def copy(self): ...
    def __hash__(self): ...

class _AssociationDict(_AssociationCollection):
    def __getitem__(self, key): ...
    def __setitem__(self, key, value) -> None: ...
    def __delitem__(self, key) -> None: ...
    def __contains__(self, key): ...
    def has_key(self, key): ...
    def __iter__(self): ...
    def clear(self) -> None: ...
    def __eq__(self, other): ...
    def __ne__(self, other): ...
    def __lt__(self, other): ...
    def __le__(self, other): ...
    def __gt__(self, other): ...
    def __ge__(self, other): ...
    def __cmp__(self, other): ...
    def get(self, key, default: Any | None = ...): ...
    def setdefault(self, key, default: Any | None = ...): ...
    def keys(self): ...
    def items(self): ...
    def values(self): ...
    def pop(self, key, default=...): ...
    def popitem(self): ...
    def update(self, *a, **kw) -> None: ...
    def copy(self): ...
    def __hash__(self): ...

class _AssociationSet(_AssociationCollection):
    def __len__(self): ...
    def __bool__(self): ...
    __nonzero__: Any
    def __contains__(self, value): ...
    def __iter__(self): ...
    def add(self, value) -> None: ...
    def discard(self, value) -> None: ...
    def remove(self, value) -> None: ...
    def pop(self): ...
    def update(self, other) -> None: ...
    def __ior__(self, other): ...  # type: ignore[misc]
    def union(self, other): ...
    __or__: Any
    def difference(self, other): ...
    __sub__: Any
    def difference_update(self, other) -> None: ...
    def __isub__(self, other): ...  # type: ignore[misc]
    def intersection(self, other): ...
    __and__: Any
    def intersection_update(self, other) -> None: ...
    def __iand__(self, other): ...  # type: ignore[misc]
    def symmetric_difference(self, other): ...
    __xor__: Any
    def symmetric_difference_update(self, other) -> None: ...
    def __ixor__(self, other): ...  # type: ignore[misc]
    def issubset(self, other): ...
    def issuperset(self, other): ...
    def clear(self) -> None: ...
    def copy(self): ...
    def __eq__(self, other): ...
    def __ne__(self, other): ...
    def __lt__(self, other): ...
    def __le__(self, other): ...
    def __gt__(self, other): ...
    def __ge__(self, other): ...
    def __hash__(self): ...



"""Tests for pytype.pyi.metadata."""

from pytype.pyi import metadata
import unittest


class StringTest(unittest.TestCase):

  def test_to_string(self):
    self.assertEqual(metadata.to_string({'tag': 'call'}),
                     "{'tag': 'call'}")

  def test_from_string(self):
    self.assertEqual(metadata.from_string("{'tag': 'call'}"),
                     {'tag': 'call'})


class CallToAnnotationTest(unittest.TestCase):

  def test_deprecated(self):
    annotation = metadata.call_to_annotation(
        'Deprecated', posargs=['squished by duck'])
    self.assertEqual(
        annotation, "{'tag': 'Deprecated', 'reason': 'squished by duck'}")

  def test_noargs(self):
    annotation = metadata.call_to_annotation('Quack')
    self.assertEqual(
        annotation,
        "{'tag': 'call', 'fn': 'Quack', 'posargs': (), 'kwargs': {}}")

  def test_posargs(self):
    annotation = metadata.call_to_annotation('Quack', posargs=[2])
    self.assertEqual(
        annotation,
        "{'tag': 'call', 'fn': 'Quack', 'posargs': [2], 'kwargs': {}}")

  def test_kwargs(self):
    annotation = metadata.call_to_annotation('Quack', kwargs={'volume': 4.5})
    self.assertEqual(annotation, ("{'tag': 'call', 'fn': 'Quack',"
                                  " 'posargs': (), 'kwargs': {'volume': 4.5}}"))

  def test_allargs(self):
    annotation = metadata.call_to_annotation(
        'Quack', posargs=[2, 'brown'],
        kwargs={'volume': 4.5, 'mode': 'correct'})
    self.assertEqual(
        annotation, ("{'tag': 'call', 'fn': 'Quack', 'posargs': [2, 'brown'],"
                     " 'kwargs': {'volume': 4.5, 'mode': 'correct'}}"))


class ToPytdTest(unittest.TestCase):

  def test_deprecated(self):
    pytd = metadata.to_pytd({'tag': 'Deprecated', 'reason': 'squished by duck'})
    self.assertEqual(pytd, "Deprecated('squished by duck')")

  def test_call_noargs(self):
    pytd = metadata.to_pytd(
        {'tag': 'call', 'fn': 'Quack', 'posargs': (), 'kwargs': {}})
    self.assertEqual(pytd, 'Quack()')

  def test_call_posargs(self):
    pytd = metadata.to_pytd(
        {'tag': 'call', 'fn': 'Quack', 'posargs': [2], 'kwargs': {}})
    self.assertEqual(pytd, 'Quack(2)')

  def test_call_kwargs(self):
    pytd = metadata.to_pytd({'tag': 'call', 'fn': 'Quack', 'posargs': (),
                             'kwargs': {'volume': 4.5}})
    self.assertEqual(pytd, 'Quack(volume=4.5)')

  def test_call_allargs(self):
    pytd = metadata.to_pytd(
        {'tag': 'call', 'fn': 'Quack', 'posargs': [2, 'brown'],
         'kwargs': {'volume': 4.5, 'mode': 'correct'}})
    self.assertEqual(pytd, "Quack(2, 'brown', volume=4.5, mode='correct')")

  def test_noncall_tag(self):
    self.assertEqual(metadata.to_pytd({'tag': 'sneeze'}), "{'tag': 'sneeze'}")

  def test_no_tag(self):
    self.assertEqual(metadata.to_pytd({}), '{}')


if __name__ == '__main__':
  unittest.main()



from typing import Any

BASE64_CHARS: Any
AB64_CHARS: Any
HASH64_CHARS: Any
BCRYPT_CHARS: Any
PADDED_BASE64_CHARS: Any
HEX_CHARS: Any
UPPER_HEX_CHARS: Any
LOWER_HEX_CHARS: Any
ALL_BYTE_VALUES: Any

def compile_byte_translation(mapping, source: Any | None = ...): ...
def b64s_encode(data): ...
def b64s_decode(data): ...
def ab64_encode(data): ...
def ab64_decode(data): ...
def b32encode(source): ...
def b32decode(source): ...

class Base64Engine:
    bytemap: Any
    big: Any
    def __init__(self, charmap, big: bool = ...) -> None: ...
    @property
    def charmap(self): ...
    def encode_bytes(self, source): ...
    def decode_bytes(self, source): ...
    def check_repair_unused(self, source): ...
    def repair_unused(self, source): ...
    def encode_transposed_bytes(self, source, offsets): ...
    def decode_transposed_bytes(self, source, offsets): ...
    def decode_int6(self, source): ...
    def decode_int12(self, source): ...
    def decode_int24(self, source): ...
    def decode_int30(self, source): ...
    def decode_int64(self, source): ...
    def encode_int6(self, value): ...
    def encode_int12(self, value): ...
    def encode_int24(self, value): ...
    def encode_int30(self, value): ...
    def encode_int64(self, value): ...

class LazyBase64Engine(Base64Engine):
    def __init__(self, *args, **kwds) -> None: ...
    def __getattribute__(self, attr): ...

h64: Any
h64big: Any
bcrypt64: Any



from typing import Any

def to_unicode(obj: float | bytes | str, encoding: str | None = ..., from_server: bool = ...) -> str: ...
def to_raw(obj, encoding: str = ...): ...
def escape_filter_chars(text: float | bytes | str, encoding: str | None = ...) -> str: ...
def unescape_filter_chars(text, encoding: Any | None = ...): ...
def escape_bytes(bytes_value: str | bytes) -> str: ...
def prepare_for_stream(value): ...
def json_encode_b64(obj): ...
def check_json_dict(json_dict) -> None: ...
def json_hook(obj): ...
def format_json(obj, iso_format: bool = ...): ...
def is_filter_escaped(text): ...
def ldap_escape_to_bytes(text): ...



from typing import Any

from ._compat_py3k import asynccontextmanager as asynccontextmanager
from ._concurrency_py3k import (
    AsyncAdaptedLock as AsyncAdaptedLock,
    await_fallback as await_fallback,
    await_only as await_only,
    greenlet_spawn as greenlet_spawn,
    is_exit_exception as is_exit_exception,
)

have_greenlet: bool
asyncio: Any | None
